# Copyright contributors to the Terratorch project

import importlib.util
import itertools
import json
import logging
import os
import shutil
import sys
import glob
import tempfile
import warnings
from copy import deepcopy
from datetime import timedelta
from pathlib import Path
from typing import Any, Optional

import albumentations
import cv2  # noqa: F401
import numpy as np
import rasterio
import torch

import random
import string

# Allows classes to be referenced using only the class name
import torchgeo.datamodules
import yaml
from albumentations.pytorch import ToTensorV2  # noqa: F401
from jsonargparse import set_dumper
from jsonargparse._namespace import Namespace
from lightning.fabric.utilities.cloud_io import get_filesystem
from lightning.fabric.utilities.types import _PATH
from lightning.pytorch import LightningDataModule, LightningModule, Trainer
from lightning.pytorch.callbacks import BasePredictionWriter, ModelCheckpoint, RichProgressBar
from lightning.pytorch.cli import ArgsType, LightningArgumentParser, LightningCLI, SaveConfigCallback
from torchgeo.trainers import BaseTask
from torch.utils.data import DataLoader
from tqdm import tqdm

import terratorch.datamodules
from terratorch.utils import compute_mask_statistics, compute_statistics
import terratorch.tasks  # noqa: F401
from terratorch.datamodules import (  # noqa: F401
    GenericNonGeoClassificationDataModule,
    GenericNonGeoPixelwiseRegressionDataModule,
    GenericNonGeoSegmentationDataModule,
)
from terratorch.datasets.transforms import (
    FlattenTemporalIntoChannels,  # noqa: F401
    Rearrange,  # noqa: F401
    SelectBands,  # noqa: F401
    UnflattenTemporalFromChannels,  # noqa: F401
)
from terratorch.datasets.utils import HLSBands

# GenericNonGeoRegressionDataModule,
from terratorch.models import PrithviModelFactory  # noqa: F401
from terratorch.models.model import AuxiliaryHead  # noqa: F401
from terratorch.tasks import (
    ClassificationTask,  # noqa: F401
    PixelwiseRegressionTask,  # noqa: F401
    SemanticSegmentationTask,  # noqa: F401
)

logger = logging.getLogger("terratorch")

from terratorch.utils import remove_unexpected_prefix

def flatten(list_of_lists):
    return list(itertools.chain.from_iterable(list_of_lists))


def open_tiff(fname):
    with rasterio.open(fname, "r") as src:
        data = src.read()
        metadata = src.meta
    return data, metadata


def is_one_band(img):
    return len(img.shape) == 2  # noqa: PLR2004


def write_tiff(img_wrt, filename, metadata):

    # Adapting the number of bands to be compatible with the 
    # output dimensions.
    if not is_one_band(img_wrt):
        count = img_wrt.shape[0]
        metadata['count'] = count

    with rasterio.open(filename, "w", **metadata) as dest:
        if is_one_band(img_wrt):
            img_wrt = img_wrt[None]

        for i in range(img_wrt.shape[0]):
            dest.write(img_wrt[i, :, :], i + 1)
    return filename

def add_default_checkpointing_config(config):
    subcommand = config.get("subcommand", None)
    if subcommand is not None:
        enable_checkpointing = config[subcommand + ".trainer.enable_checkpointing"]
        callbacks = config[subcommand + ".trainer.callbacks"]

        # A list for callbacks is usually expected.
        if callbacks:
            check_callbacks = [op for op in callbacks if "ModelCheckpoint" in op.class_path]

            if len(check_callbacks) > 0:
                there_is_checkpointing = True
            else:
                there_is_checkpointing = False
        # However, if no callbacks list is provided, let's create one. 
        else:
            there_is_checkpointing = False
            callbacks = []

        if enable_checkpointing:
            if not there_is_checkpointing:
                # Checkpointing is enabled, but no checkpoint rule was
                # provided, so let's define some default choice for it.
                model_checkpoint = Namespace(class_path="StateDictAwareModelCheckpoint",
                                            init_args=Namespace(filename="{epoch}", monitor="val/loss"))
                model_checkpoint_state_dict = Namespace(class_path="StateDictAwareModelCheckpoint",
                                                   init_args=Namespace(filename="{epoch}_state_dict",
                                                                       save_weights_only=True, monitor="val/loss"))
                callbacks += [model_checkpoint, model_checkpoint_state_dict]   
                config[subcommand + ".trainer.callbacks"] = callbacks
            else:
                logger.info("No extra checkpoint config will be added, since the user already defined it in the callbacks.")

    return config 

def save_prediction(prediction, input_file_name, out_dir, dtype:str="int16",
                    suffix:str="pred", output_file_name:str | None=None):

    mask, metadata = open_tiff(input_file_name)
    mask = np.where(mask == metadata["nodata"], 1, 0)
    mask = np.max(mask, axis=0)
    result = np.where(mask == 1, -1, prediction.detach().cpu())

    ##### Save file to disk
    metadata["count"] = 1
    metadata["dtype"] = dtype
    metadata["compress"] = "lzw"
    metadata["nodata"] = -1
    file_name = os.path.basename(input_file_name)

    if not output_file_name:
        file_name_no_ext = os.path.splitext(file_name)[0]
    else:
        file_name_no_ext = output_file_name

    out_file_name = file_name_no_ext + f"_{suffix}.tif"
    logger.info(f"Saving output to {out_file_name} ...")
    write_tiff(result, os.path.join(out_dir, out_file_name), metadata)


def import_custom_modules(custom_modules_path: str | Path | None = None) -> None:

    if custom_modules_path:

        custom_modules_path = Path(custom_modules_path)

        if custom_modules_path.is_dir():

            # Add 'custom_modules' folder to sys.path
            workdir = custom_modules_path.parents[0]
            module_dir = custom_modules_path.name

            sys.path.insert(0, str(workdir))

            try:
                module = importlib.import_module(module_dir)
                logger.info(f"Found {custom_modules_path}")
            except ImportError:
                raise ImportError(f"It was not possible to import modules from {custom_modules_path}.")
        else:
            raise ValueError(f"Modules path {custom_modules_path} isn't a directory. Check if you have defined it properly.")
    else:
        logger.debug("No custom module is being used.")

class CustomWriter(BasePredictionWriter):
    """Callback class to write geospatial data to file."""

    def __init__(self, output_dir: str | None = None, write_interval: str = "epoch"):

        super().__init__(write_interval)

        self.output_dir = output_dir

    def write_on_batch_end(self, trainer, pl_module, prediction, batch_indices, batch, batch_idx, dataloader_idx):  # noqa: ARG002
        # this will create N (num processes) files in `output_dir` each containing
        # the predictions of it's respective rank

        # by default take self.output_dir. If None, look for one in trainer
        if self.output_dir is None:
            try:
                output_dir = trainer.predict_output_dir
            except AttributeError as err:
                msg = "Output directory must be passed to CustomWriter constructor or the `predict_output_dir`\
                attribute must be present in the trainer."
                raise Exception(msg) from err
        else:
            output_dir = self.output_dir

        # Defining a prefix for the output files
        if hasattr(trainer, "output_file_prefix"):
            output_file_prefix = trainer.output_file_prefix
        else:
            output_file_prefix = None
            msg = "As you don't defined a value for `--output_file_prefix` via comand line, we will use a name based on the input file names."
            logger.info(msg)

        if not os.path.exists(output_dir):
            os.makedirs(output_dir, exist_ok=True)

        if isinstance(prediction, torch.Tensor):
            filename_batch = ''.join(random.choices(string.ascii_letters + string.digits, k=8))
            torch.save(prediction, os.path.join(output_dir, f"{filename_batch}.pt"))

        elif isinstance(prediction, tuple):

            pred_batch_, filename_batch = prediction

            # In case of multimodal datasets, we have a dictionary
            # with one file per modality. We need to use a file as 
            # input in order to copy metadata. 
            if type(filename_batch)==dict:
                keys = list(filename_batch.keys())
                filename_batch = filename_batch[keys[0]]

            # If we have just a single kind of output
            if isinstance(pred_batch_, tuple):

                pred_batch, suffix = pred_batch_

                for prediction, file_name in zip(torch.unbind(pred_batch, dim=0), filename_batch, strict=False):
                    save_prediction(prediction, file_name, output_dir,
                                    dtype=trainer.out_dtype, suffix=suffix,
                                    output_file_name=output_file_prefix)

            # If we are outputting more than one kind of variable, as
            # (predictions, probabilities), for segmentation
            elif isinstance(pred_batch_, list):

                for pred in pred_batch_:
                    pred_batch, suffix = pred

                    for p, file_name in zip(torch.unbind(pred_batch, dim=0), filename_batch, strict=False):
                        save_prediction(p, file_name, output_dir,
                                        dtype=trainer.out_dtype, suffix=suffix,
                                        output_file_name=output_file_prefix)

        else:
            raise TypeError(f"Unknown type for prediction {type(prediction)}")

    def write_on_epoch_end(self, trainer, pl_module, predictions, batch_indices):  # noqa: ARG002
        # this will create N (num processes) files in `output_dir` each containing
        # the predictions of it's respective rank

        # by default take self.output_dir. If None, look for one in trainer
        if self.output_dir is None:
            try:
                output_dir = trainer.predict_output_dir
            except AttributeError as err:
                msg = "Output directory must be passed to CustomWriter constructor or the `predict_output_dir`\
                attribute must be present in the trainer."
                raise Exception(msg) from err
        else:
            output_dir = self.output_dir

        if not os.path.exists(output_dir):
            os.makedirs(output_dir, exist_ok=True)

        for pred_batch, filename_batch in predictions:
            for prediction, file_name in zip(torch.unbind(pred_batch, dim=0), filename_batch, strict=False):
                save_prediction(prediction, file_name, output_dir, dtype=trainer.out_dtype)


def clean_config_for_deployment_and_dump(config: dict[str, Any]):
    deploy_config = deepcopy(config)
    ## General
    # drop ckpt_path
    deploy_config.pop("ckpt_path", None)
    # drop checkpoints
    deploy_config.pop("ModelCheckpoint", None)
    deploy_config.pop("StateDictModelCheckpoint", None)
    # drop optimizer and lr sheduler
    deploy_config.pop("optimizer", None)
    deploy_config.pop("lr_scheduler", None)
    # drop undesirable fields, if they exist
    if "verbose" in deploy_config:
        deploy_config.pop("verbose", None)
    ## Trainer
    # remove logging
    deploy_config["trainer"]["logger"] = False
    # remove callbacks
    deploy_config["trainer"].pop("callbacks", None)
    # remove default_root_dir
    deploy_config["trainer"].pop("default_root_dir", None)
    # set mixed precision by default for inference
    deploy_config["trainer"]["precision"] = "16-mixed"
    ## Model
    # set pretrained to false
    if "model_args" in deploy_config["model"]["init_args"]:
        # for the new prithvi EO v2 models, alter the backbone_pretrained value only.
        if "pretrained" in deploy_config["model"]["init_args"]["model_args"]:
            deploy_config["model"]["init_args"]["model_args"]["pretrained"] = False
        elif "backbone_pretrained" in deploy_config["model"]["init_args"]["model_args"]:
            deploy_config["model"]["init_args"]["model_args"]["backbone_pretrained"] = False
            
    # Set image_grep and label_grep to *tif* . 
    # Fixes issue with inference not finding image named as the training image.
    deploy_config['data']['init_args']['img_grep'] = "*tif*"
    deploy_config['data']['init_args']['label_grep'] = "*tif*"
    return yaml.safe_dump(deploy_config)


class StudioDeploySaveConfigCallback(SaveConfigCallback):
    """Subclass of the SaveConfigCallback to save configs.
    Besides saving the usual config, will also save a modified config ready to be uploaded.
    Handles things like disabling the pretrained flag, removing loggers, ...
    """

    def __init__(
        self,
        parser,
        config,
        config_filename: str = "config.yaml",
        overwrite: bool = False,
        multifile: bool = False,
        save_to_log_dir: bool = True,
    ):
        super().__init__(parser, config, config_filename, overwrite, multifile, save_to_log_dir)
        set_dumper("deploy_config", clean_config_for_deployment_and_dump)

        # Preparing information to save config file to log dir
        config_dict = config.as_dict()
        self.config_path_original = str(config_dict["config"][0])
        _, self.config_file_original = os.path.split(self.config_path_original)

        self.deploy_config_file = config_dict["deploy_config_file"]

    def setup(self, trainer: Trainer, pl_module: LightningModule, stage: str) -> None:
        if self.already_saved:
            return

        if self.save_to_log_dir:
            log_dir = trainer.log_dir or trainer.default_root_dir  # this broadcasts the directory
            if log_dir is None:
                msg = "log_dir was none. Please set a logger or set default_root_dir for the trainer"
                raise Exception(msg)
            config_path = os.path.join(log_dir, self.config_filename)
            fs = get_filesystem(log_dir)

            if not self.overwrite:
                # check if the file exists on rank 0
                file_exists = fs.isfile(config_path) if trainer.is_global_zero else False
                # broadcast whether to fail to all ranks
                file_exists = trainer.strategy.broadcast(file_exists)
                if file_exists:
                    msg = (
                        f"{self.__class__.__name__} expected {config_path} to NOT exist. Aborting to avoid overwriting"
                        " results of a previous run. You can delete the previous config file,"
                        " set `LightningCLI(save_config_callback=None)` to disable config saving,"
                        ' or set `LightningCLI(save_config_kwargs={"overwrite": True})` to overwrite the config file.'
                    )
                    raise RuntimeError(msg)

            if trainer.is_global_zero:
                # save only on rank zero to avoid race conditions.
                # the `log_dir` needs to be created as we rely on the logger to do it usually
                # but it hasn't logged anything at this point
                fs.makedirs(log_dir, exist_ok=True)

                if self.deploy_config_file:
                    self.parser.save(
                        self.config, config_path, skip_none=True, overwrite=self.overwrite, multifile=self.multifile
                    )

        if trainer.is_global_zero:
            if self.deploy_config_file:
                # also save the config that will be deployed
                config_name, config_ext = os.path.splitext(self.config_filename)
                config_name += "_deploy"
                config_name += config_ext
                config_path = os.path.join(log_dir, config_name)
                self.parser.save(
                    self.config,
                    config_path,
                    format="deploy_config",
                    skip_none=True,
                    overwrite=self.overwrite,
                    multifile=self.multifile,
                )
                self.already_saved = True

        config_path_dir, config_path_file = os.path.split(config_path)
        self.config_path_new = os.path.join(config_path_dir, self.config_file_original)

        # broadcast so that all ranks are in sync on future calls to .setup()
        self.already_saved = trainer.strategy.broadcast(self.already_saved)
        # Copying config file to log dir only if they are different files
        if os.path.abspath(self.config_path_original) != os.path.abspath(self.config_path_new):
            shutil.copyfile(self.config_path_original, self.config_path_new)

class StateDictAwareModelCheckpoint(ModelCheckpoint):
    # necessary as we wish to have one model checkpoint with only state dict and one with standard lightning checkpoints
    # for this, the state key needs to be different, and thus to include this save_weights_only parameter
    def __init__(
        self,
        dirpath: _PATH | None = None,
        filename: str | None = None,
        monitor: str | None = None,
        verbose: bool = False,
        save_last: bool | None = None,
        save_top_k: int = 1,
        save_best_only: bool = False,
        mode: str = "min",
        save_weights_only: bool = False,
        auto_insert_metric_name: bool = True,
        every_n_train_steps: int | None = None,
        train_time_interval: timedelta | None = None,
        every_n_epochs: int | None = None,
        save_on_train_epoch_end: bool | None = None,
        enable_version_counter: bool = True,
    ):
        if save_best_only:
            save_top_k = 1

        super().__init__(
            dirpath,
            filename,
            monitor,
            verbose,
            save_last,
            save_top_k,
            save_weights_only,
            mode,
            auto_insert_metric_name,
            every_n_train_steps,
            train_time_interval,
            every_n_epochs,
            save_on_train_epoch_end,
            enable_version_counter,
        )

    @property
    def state_key(self) -> str:
        return self._generate_state_key(
            monitor=self.monitor,
            mode=self.mode,
            every_n_train_steps=self._every_n_train_steps,
            every_n_epochs=self._every_n_epochs,
            train_time_interval=self._train_time_interval,
            save_weights_only=self.save_weights_only,
        )


class MyLightningCLI(LightningCLI):
    def run_init(self):
        logger.info("Running custom init command...")

    @property
    def subcommands(self):
        return super().subcommands + ["init"]

    def run(self):
        subcommand = self.config['subcommand']
        if subcommand == 'init':
            self.run_init()
        else:
            super().run()

    def add_arguments_to_parser(self, parser: LightningArgumentParser) -> None:
        parser.add_argument("--predict_output_dir", default=None)
        parser.add_argument("--output_file_prefix", default=None)
        parser.add_argument("--out_dtype", default="int16")
        parser.add_argument("--deploy_config_file", type=bool, default=True)
        parser.add_argument("--custom_modules_path", type=str, default=None)

    def instantiate_classes(self) -> None:

        # Adding default configuration for checkpoint saving when 
        # enable_checkpointing is True and no checkpointing is included as
        # callback. 
        self.config = add_default_checkpointing_config(self.config)

        # get the predict_output_dir. Depending on the value of run, it may be in the subcommand
        try:
            config = self.config.predict
        except AttributeError:
            config = self.config

        # Custom modules path
        if hasattr(self.config, "fit") and hasattr(self.config.fit, "custom_modules_path"):
            custom_modules_path = self.config.fit.custom_modules_path
        elif hasattr(self.config, "validate") and hasattr(self.config.validate, "custom_modules_path"):
            custom_modules_path = self.config.validate.custom_modules_path
        elif hasattr(self.config, "test") and hasattr(self.config.test, "custom_modules_path"):
            custom_modules_path = self.config.test.custom_modules_path
        elif hasattr(self.config, "predict") and hasattr(self.config.predict, "custom_modules_path"):
            custom_modules_path = self.config.predict.custom_modules_path
        else:
            custom_modules_path = os.getenv("TERRATORCH_CUSTOM_MODULE_PATH", None)

        import_custom_modules(custom_modules_path)

        super().instantiate_classes()

        if hasattr(config, "predict_output_dir"):
            self.trainer.predict_output_dir = config.predict_output_dir

        if hasattr(config, "output_file_prefix"):
            self.trainer.output_file_prefix = config.output_file_prefix

        if hasattr(config, "out_dtype"):
            self.trainer.out_dtype = config.out_dtype

        if hasattr(config, "deploy_config_file"):
            self.trainer.deploy_config = config.deploy_config_file

    @staticmethod
    def subcommands() -> dict[str, set[str]]:
        existing_subcommands = LightningCLI.subcommands()
        existing_subcommands["compute_statistics"] = {"datamodule"}
        return existing_subcommands


def build_lightning_cli(
    args: ArgsType = None,
    run=True,  # noqa: FBT002
) -> LightningCLI:
    """Command-line interface to TerraTorch."""
    # Taken from https://github.com/pangeo-data/cog-best-practices
    rasterio_best_practices = {
        "GDAL_DISABLE_READDIR_ON_OPEN": "EMPTY_DIR",
        "AWS_NO_SIGN_REQUEST": "YES",
        "GDAL_MAX_RAW_BLOCK_CACHE_SIZE": "200000000",
        "GDAL_SWATH_SIZE": "200000000",
        "VSI_CURL_CACHE_SIZE": "200000000",
    }
    os.environ.update(rasterio_best_practices)
    if os.environ.get("terratorch_FLOAT_32_PRECISION") is not None:
        precision = os.environ.get("terratorch_FLOAT_32_PRECISION")
        allowed_values = ["highest", "high", "medium"]
        if precision in allowed_values:
            torch.set_float32_matmul_precision(precision)
        else:
            warnings.warn(
                "Found terratorch_FLOAT_32_PRECISION env variable but value was set to precision.\
                Set to one of {allowed_values}. Will be ignored this run.",
                UserWarning,
                stacklevel=1,
            )

    return MyLightningCLI(
        model_class=BaseTask,
        subclass_mode_model=True,
        subclass_mode_data=True,
        seed_everything_default=0,
        save_config_callback=StudioDeploySaveConfigCallback if run else None,
        save_config_kwargs={"overwrite": True},
        args=args,
        # save only state_dict as well as full state. Only state_dict will be used for exporting the model
        trainer_defaults={"callbacks": [CustomWriter(write_interval="batch")]},
        run=run,
        trainer_class=MyTrainer,
    )


class LightningInferenceModel:
    """This class provides all the same functionalities as loading a model from the CLI,
    but makes the model accessible programatically.

    Example usage:
        model = LightningInferenceModel.from_config(
        <config path>,
        <checkpoint path>,
        <output dir path>,
        )
        model.inference(<path to inference data dir>)
    """

    def __init__(
        self,
        trainer: Trainer,
        model: LightningModule,
        datamodule: LightningDataModule,
        checkpoint_path: Path | None = None,
    ):
        self.trainer = trainer
        self.model = model
        self.datamodule = datamodule
        if checkpoint_path:
            weights = torch.load(checkpoint_path, map_location="cpu", weights_only=True)
            if "state_dict" in weights:
                weights = weights["state_dict"]

            # It removes a residual prefix (related to timm) from older
            # checkpoints.
            weights = remove_unexpected_prefix(weights)
            # Removing the undesirable prefix `model` from the checkpoint
            weights_ = {}
            for k, v in weights.items():
                splits = k.split(".")
                if splits[0] == "model":
                    splits = splits[1:]
                    k_ = ".".join(splits)
                    weights_[k_] = v
                else:
                    weights_[k] = v

            self.model.model.load_state_dict(weights_)

        # dont write
        non_writing_callbacks = []
        for callback in self.trainer.callbacks:
            if isinstance(callback, BasePredictionWriter):
                continue
            non_writing_callbacks.append(callback)
        self.trainer.callbacks = non_writing_callbacks

    @staticmethod
    def from_config(
        config_path: Path,
        checkpoint_path: Path | None = None,
        predict_dataset_bands: list[str] | None = None,
        predict_output_bands: list[str] | None = None,
    ):
        """
        Args:
            config_path (Path): Path to the config of the model to be loaded.
            checkpoint_path (Path): Path to the checkpoint to be loaded.
            predict_dataset_bands (list[str] | None, optional): List of bands present in input data.
                Defaults to None.
        """
        # use cli only to load
        arguments = [
            "--config",
            config_path,
        ]

        if predict_dataset_bands is not None:
            arguments.extend([ "--data.init_args.predict_dataset_bands",
            "[" + ",".join(predict_dataset_bands) + "]",])

        if predict_output_bands is not None:
            arguments.extend([ "--data.init_args.predict_output_bands",
            "[" + ",".join(predict_output_bands) + "]",])

        cli = build_lightning_cli(arguments, run=False)
        trainer = cli.trainer
        # disable logging metrics
        trainer.logger = None
        datamodule = cli.datamodule
        model = cli.model
        return LightningInferenceModel(trainer, model, datamodule, checkpoint_path=checkpoint_path)

    @staticmethod
    def from_task_and_datamodule(
        task: LightningModule,
        datamodule: LightningDataModule,
        checkpoint_path: Path | None = None,
    ):
        trainer = Trainer(
            # precision="16-mixed",
            callbacks=[RichProgressBar()],
        )
        return LightningInferenceModel(trainer, task, datamodule, checkpoint_path=checkpoint_path)

    def inference_on_dir(self, data_root: Path | None = None) -> tuple[torch.Tensor, list[str]]:
        """Perform inference on the given data root directory

        Args:
            data_root (Path): Directory with data to perform inference on
        Returns:
            A tuple with a torch tensor with all predictions and a list of corresponding input file paths

        """

        if data_root:
            self.datamodule.predict_root = data_root
        predictions = self.trainer.predict(model=self.model, datamodule=self.datamodule, return_predictions=True)

        # In some cases, the output has the format ((prediction_tensor,
        # prediction_name), filename)
        if all([type(j[0])==tuple for j in predictions]):
            concat_predictions = torch.cat([batch[0][0] for batch in predictions])
        else:
            concat_predictions = torch.cat([batch[0] for batch in predictions])

        concat_file_names = flatten([batch[1] for batch in predictions])

        return concat_predictions, concat_file_names

    def inference(self, file_path: Path) -> torch.Tensor:
        """Perform inference on a single input file path

        Args:
            file_path (Path): Path to a single input file.
        Returns:
            A torch tensor with the prediction
        """

        with tempfile.TemporaryDirectory() as tmpdir:
            os.symlink(file_path, os.path.join(tmpdir, os.path.basename(file_path)))
            prediction, file_name = self.inference_on_dir(
                tmpdir,
            )
            return prediction.squeeze(0)


class MyTrainer(Trainer):
    def compute_statistics(self, datamodule: LightningDataModule, **kwargs) -> None:
        """
        Compute the dataset statistics for the training dataset.

        This method will compute the mean and standard deviation of the image data and the count and percentage of each
        unique value in the masks in case these are int and the mean and standard deviation of the mask values in case
        these are floats. The statistics are computed using the entire training dataset and are printed to the logger.

        Please note that this method assumes that there is only one train dataloader in the datamodule. The train
        transforms are removed before computing the statistics to ensure that the statistics are computed on the raw
        data without any augmentation and randomization.
        """
        # remove train transforms, this may not work for all datamodules
        if hasattr(datamodule, "train_transform"):
            datamodule.train_transform = None
        datamodule.setup("fit")
        original_dataloader = datamodule.train_dataloader()
        if not isinstance(original_dataloader, DataLoader):
            msg = "DataLoader not found in datamodule.train_dataloader()"
            raise ValueError(msg)
        new_dataloader = DataLoader(
            dataset=original_dataloader.dataset,
            batch_size=original_dataloader.batch_size,
            shuffle=False,
            num_workers=original_dataloader.num_workers,
            collate_fn=original_dataloader.collate_fn,
            pin_memory=original_dataloader.pin_memory,
            drop_last=False,
        )
        image_stats = compute_statistics(new_dataloader)
        logger.info("Image statistics:")
        logger.info(yaml.dump(image_stats))
        if "mask" in datamodule.train_dataloader().dataset[0]:
            mask_stats = compute_mask_statistics(new_dataloader)
            logger.info("Mask statistics:")
            logger.info(yaml.dump(mask_stats))



# Copyright 2025 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
from torch import nn
import torch.nn.functional as F

from einops import rearrange, pack, unpack


def pack_one(t, pattern):
    return pack([t], pattern)


def unpack_one(t, ps, pattern):
    return unpack(t, ps, pattern)[0]


def round_ste(x):
    """Round with straight through gradients."""
    xhat = x.round()
    return x + (xhat - x).detach()


class FiniteScalarQuantizer(nn.Module):
    def __init__(
        self,
        codebook_size: str,  # example: "8-8-8-6-5"
    ):
        super().__init__()

        levels = [int(level) for level in codebook_size.split('-')]

        _levels = torch.tensor(levels, dtype=torch.int32)
        self.register_buffer("_levels", _levels, persistent=False)
        _basis = torch.cumprod(
            torch.tensor([1] + levels[:-1]), dim=0, dtype=torch.int32
        )
        self.register_buffer("_basis", _basis, persistent=False)

        # initialize codebook
        self.codebook_size = self._levels.prod().item()
        codebook = self.indice_to_code(torch.arange(self.codebook_size))
        self.register_buffer("codebook", codebook, persistent=False)

    def latent_to_code_and_indice(self, latent):
        d = self._levels - 1
        number = round_ste(F.sigmoid(latent) * d)
        code = number / d
        indice = (number * self._basis).sum(dim=-1).to(torch.int32)

        return code, indice

    def indice_to_code(self, indice):
        # (..., d)
        code = (indice.unsqueeze(-1) // self._basis) % self._levels
        # convert to [0, 1]
        code = code / (self._levels - 1)

        return code

    def indices_to_embedding(self, indices):
        B, H, W = indices.shape
        indices = rearrange(indices, "b h w -> b (h w)")
        embeddings = self.indice_to_code(indices)
        embeddings = rearrange(embeddings, "b (h w) c -> b c h w", h=H)
        return embeddings

    def forward(self, x):
        height, width = x.shape[-2:]
        x = rearrange(x, 'b c h w -> b (h w) c')

        quantize, embed_ind = self.latent_to_code_and_indice(x)

        quantize = rearrange(quantize, 'b (h w) c -> b c h w', h=height, w=width)
        embed_ind = rearrange(embed_ind, 'b (h w) ... -> b h w ...', h=height, w=width)

        # no auxiliary losses needed for FSQ
        loss = torch.tensor([0.0], device=x.device, requires_grad=self.training)

        return quantize, loss, embed_ind
    


from collections.abc import Sequence
from typing import Any

import albumentations as A
from kornia.augmentation import AugmentationSequential

from terratorch.datamodules.geobench_data_module import GeobenchDataModule
from terratorch.datasets import MBeninSmallHolderCashewsNonGeo

MEANS = {
    "COASTAL_AEROSOL": 520.1185302734375,
    "BLUE": 634.7583618164062,
    "GREEN": 892.461181640625,
    "RED": 880.7075805664062,
    "RED_EDGE_1": 1380.6409912109375,
    "RED_EDGE_2": 2233.432373046875,
    "RED_EDGE_3": 2549.379638671875,
    "NIR_BROAD": 2643.248046875,
    "NIR_NARROW": 2643.531982421875,
    "WATER_VAPOR": 2852.87451171875,
    "SWIR_1": 2463.933349609375,
    "SWIR_2": 1600.9207763671875,
    "CLOUD_PROBABILITY": 0.010281000286340714,
}

STDS = {
    "COASTAL_AEROSOL": 204.2023468017578,
    "BLUE": 227.25344848632812,
    "GREEN": 222.32545471191406,
    "RED": 350.47235107421875,
    "RED_EDGE_1": 280.6436767578125,
    "RED_EDGE_2": 373.7521057128906,
    "RED_EDGE_3": 449.9236145019531,
    "NIR_BROAD": 414.6498107910156,
    "NIR_NARROW": 415.1019592285156,
    "WATER_VAPOR": 413.8980407714844,
    "SWIR_1": 494.97430419921875,
    "SWIR_2": 514.4229736328125,
    "CLOUD_PROBABILITY": 0.3447800576686859,
}


class MBeninSmallHolderCashewsNonGeoDataModule(GeobenchDataModule):
    """NonGeo LightningDataModule implementation for M-Cashew Plantation dataset."""

    def __init__(
        self,
        batch_size: int = 8,
        num_workers: int = 0,
        data_root: str = "./",
        bands: Sequence[str] | None = None,
        train_transform: A.Compose | None | list[A.BasicTransform] = None,
        val_transform: A.Compose | None | list[A.BasicTransform] = None,
        test_transform: A.Compose | None | list[A.BasicTransform] = None,
        aug: AugmentationSequential = None,
        partition: str = "default",
        use_metadata: bool = False,  # noqa: FBT002, FBT001
        **kwargs: Any,
    ) -> None:
        """
        Initializes the MBeninSmallHolderCashewsNonGeoDataModule for the M-BeninSmallHolderCashewsNonGeo dataset.

        Args:
            batch_size (int, optional): Batch size for DataLoaders. Defaults to 8.
            num_workers (int, optional): Number of workers for data loading. Defaults to 0.
            data_root (str, optional): Root directory of the dataset. Defaults to "./".
            bands (Sequence[str] | None, optional): List of bands to use. Defaults to None.
            train_transform (A.Compose | None | list[A.BasicTransform], optional): Transformations for training.
            val_transform (A.Compose | None | list[A.BasicTransform], optional): Transformations for validation.
            test_transform (A.Compose | None | list[A.BasicTransform], optional): Transformations for testing.
            aug (AugmentationSequential, optional): Augmentation/normalization pipeline. Defaults to None.
            partition (str, optional): Partition size. Defaults to "default".
            use_metadata (bool): Whether to return metadata info.
            **kwargs (Any): Additional keyword arguments.
        """
        super().__init__(
            MBeninSmallHolderCashewsNonGeo,
            MEANS,
            STDS,
            batch_size=batch_size,
            num_workers=num_workers,
            data_root=data_root,
            bands=bands,
            train_transform=train_transform,
            val_transform=val_transform,
            test_transform=test_transform,
            aug=aug,
            partition=partition,
            use_metadata=use_metadata,
            **kwargs,
        )



from collections.abc import Sequence
from typing import Any

import albumentations as A
from kornia.augmentation import AugmentationSequential

from terratorch.datamodules.geobench_data_module import GeobenchDataModule
from terratorch.datasets import MSo2SatNonGeo

MEANS = {
    "VH_REAL": 0.00030114364926703274,
    "BLUE": 0.12951652705669403,
    "VH_IMAGINARY": -5.6475887504348066e-06,
    "GREEN": 0.11734361201524734,
    "VV_REAL": 0.000289927760604769,
    "RED": 0.11374464631080627,
    "VV_IMAGINARY": -0.0005758664919994771,
    "VH_LEE_FILTERED": 0.056090500205755234,
    "RED_EDGE_1": 0.12693354487419128,
    "VV_LEE_FILTERED": 0.33695536851882935,
    "RED_EDGE_2": 0.16917912662029266,
    "VH_LEE_FILTERED_REAL": -0.0026130808982998133,
    "RED_EDGE_3": 0.19080990552902222,
    "NIR_BROAD": 0.18381330370903015,
    "VV_LEE_FILTERED_IMAGINARY": -0.00021384705905802548,
    "NIR_NARROW": 0.20517952740192413,
    "SWIR_1": 0.1762811541557312,
    "SWIR_2": 0.1286638230085373,
}

STDS = {
    "VH_REAL": 0.20626230537891388,
    "BLUE": 0.040680479258298874,
    "VH_IMAGINARY": 0.19834314286708832,
    "GREEN": 0.05125178396701813,
    "VV_REAL": 0.5187134146690369,
    "RED": 0.07254913449287415,
    "VV_IMAGINARY": 0.519291877746582,
    "VH_LEE_FILTERED": 1.8058000802993774,
    "RED_EDGE_1": 0.06872648745775223,
    "VV_LEE_FILTERED": 4.893222808837891,
    "RED_EDGE_2": 0.07402216643095016,
    "VH_LEE_FILTERED_REAL": 0.8903943300247192,
    "RED_EDGE_3": 0.08412779122591019,
    "NIR_BROAD": 0.08534552156925201,
    "VV_LEE_FILTERED_IMAGINARY": 1.2786953449249268,
    "NIR_NARROW": 0.09248979389667511,
    "SWIR_1": 0.10270608961582184,
    "SWIR_2": 0.09284552931785583,
}


class MSo2SatNonGeoDataModule(GeobenchDataModule):
    """NonGeo LightningDataModule implementation for M-So2Sat dataset."""

    def __init__(
        self,
        batch_size: int = 8,
        num_workers: int = 0,
        data_root: str = "./",
        bands: Sequence[str] | None = None,
        train_transform: A.Compose | None | list[A.BasicTransform] = None,
        val_transform: A.Compose | None | list[A.BasicTransform] = None,
        test_transform: A.Compose | None | list[A.BasicTransform] = None,
        aug: AugmentationSequential = None,
        partition: str = "default",
        **kwargs: Any,
    ) -> None:
        """
        Initializes the MSo2SatNonGeoDataModule for the MSo2SatNonGeo dataset.

        Args:
            batch_size (int, optional): Batch size for DataLoaders. Defaults to 8.
            num_workers (int, optional): Number of workers for data loading. Defaults to 0.
            data_root (str, optional): Root directory of the dataset. Defaults to "./".
            bands (Sequence[str] | None, optional): List of bands to use. Defaults to None.
            train_transform (A.Compose | None | list[A.BasicTransform], optional): Transformations for training.
            val_transform (A.Compose | None | list[A.BasicTransform], optional): Transformations for validation.
            test_transform (A.Compose | None | list[A.BasicTransform], optional): Transformations for testing.
            aug (AugmentationSequential, optional): Augmentation/normalization pipeline. Defaults to None.
            partition (str, optional): Partition size. Defaults to "default".
            **kwargs (Any): Additional keyword arguments.
        """
        super().__init__(
            MSo2SatNonGeo,
            MEANS,
            STDS,
            batch_size=batch_size,
            num_workers=num_workers,
            data_root=data_root,
            bands=bands,
            train_transform=train_transform,
            val_transform=val_transform,
            test_transform=test_transform,
            aug=aug,
            partition=partition,
            **kwargs,
        )



# Copyright contributors to the Terratorch project
import logging
import warnings
from collections.abc import Callable

import torch
from huggingface_hub import hf_hub_download
from torch import Tensor, nn

from terratorch.datasets import HLSBands
from terratorch.datasets.utils import generate_bands_intervals
from terratorch.models.backbones.prithvi_mae import PrithviViT, PrithviMAE
from terratorch.models.backbones.prithvi_vit_adapter import PrithviViTAdapter
from terratorch.models.backbones.select_patch_embed_weights import select_patch_embed_weights
from terratorch.registry import TERRATORCH_BACKBONE_REGISTRY, TERRATORCH_FULL_MODEL_REGISTRY
from huggingface_hub import hf_hub_download


logger = logging.getLogger(__name__)

PRETRAINED_BANDS = [
    HLSBands.BLUE,
    HLSBands.GREEN,
    HLSBands.RED,
    HLSBands.NIR_NARROW,
    HLSBands.SWIR_1,
    HLSBands.SWIR_2,
]

PRITHVI_V1_MEAN = [775.0, 1081.0, 1229.0, 2497.0, 2204.0, 1611.0]
PRITHVI_V1_STD = [1282.0, 1270.0, 1399.0, 1368.0, 1292.0, 1155.0]
PRITHVI_V2_MEAN = [1087.0, 1342.0, 1433.0, 2734.0, 1958.0, 1363.0]
PRITHVI_V2_STD = [2248.0, 2179.0, 2178.0, 1850.0, 1242.0, 1049.0]


def _cfg(**kwargs):
    return {
        "img_size": 224,
        "num_frames": 4,
        "patch_size": [1, 16, 16],
        "in_chans": 6,
        "embed_dim": 768,
        "depth": 12,
        "num_heads": 12,
        "decoder_embed_dim": 512,
        "decoder_depth": 8,
        "decoder_num_heads": 16,
        "mlp_ratio": 4,
        "mean": PRITHVI_V2_MEAN,
        "std": PRITHVI_V2_STD,
        "coords_encoding": [],
        "coords_scale_learn": False,
        "bands": PRETRAINED_BANDS,
        "mask_ratio": 0.75,
        "norm_pix_loss": False,
        **kwargs,
    }


prithvi_cfgs = {
    "prithvi_eo_tiny": _cfg(
        num_frames=1, embed_dim=256, depth=4, num_heads=4, decoder_embed_dim=128, decoder_depth=4, decoder_num_heads=4
    ),
    "prithvi_eo_v1_100": _cfg(num_frames=3, mean=PRITHVI_V1_MEAN, std=PRITHVI_V1_STD),
    "prithvi_eo_v2_100": _cfg(),
    "prithvi_eo_v2_300": _cfg(embed_dim=1024, depth=24, num_heads=16),
    "prithvi_eo_v2_300_tl": _cfg(
        embed_dim=1024, depth=24, num_heads=16, coords_encoding=["time", "location"], coords_scale_learn=True
    ),
    "prithvi_eo_v2_600": _cfg(embed_dim=1280, depth=32, num_heads=16, patch_size=[1, 14, 14]),
    "prithvi_eo_v2_600_tl": _cfg(
        embed_dim=1280,
        depth=32,
        num_heads=16,
        patch_size=[1, 14, 14],
        coords_encoding=["time", "location"],
        coords_scale_learn=True,
    ),
}

pretrained_weights = {
    "prithvi_eo_v1_100": {
        "hf_hub_id": "ibm-nasa-geospatial/Prithvi-EO-1.0-100M",
        "hf_hub_filename": "Prithvi_EO_V1_100M.pt",
    },
    "prithvi_eo_v2_300": {
        "hf_hub_id": "ibm-nasa-geospatial/Prithvi-EO-2.0-300M",
        "hf_hub_filename": "Prithvi_EO_V2_300M.pt",
    },
    "prithvi_eo_v2_300_tl": {
        "hf_hub_id": "ibm-nasa-geospatial/Prithvi-EO-2.0-300M-TL",
        "hf_hub_filename": "Prithvi_EO_V2_300M_TL.pt",
    },
    "prithvi_eo_v2_600": {
        "hf_hub_id": "ibm-nasa-geospatial/Prithvi-EO-2.0-600M",
        "hf_hub_filename": "Prithvi_EO_V2_600M.pt",
    },
    "prithvi_eo_v2_600_tl": {
        "hf_hub_id": "ibm-nasa-geospatial/Prithvi-EO-2.0-600M-TL",
        "hf_hub_filename": "Prithvi_EO_V2_600M_TL.pt",
    },
}

prithvi_adapter_cfgs = {
    "prithvi_eo_v1_100": {
        "interaction_indexes": [[0, 2], [3, 5], [6, 8], [9, 11]],
        "conv_inplane": 64,
        "add_vit_feature": True,
        "deform_num_heads": 12,
        "n_points": 4,
        "init_values": 0,
        "with_cffn": False,
        "cffn_ratio": 0.25,
        "deform_ratio": 0.5,
        "use_extra_extractor": True,
        "with_cp": False,
    },
    "prithvi_eo_v2_300": {
        "interaction_indexes": [[0, 5], [6, 11], [12, 17], [18, 23]],
        "conv_inplane": 64,
        "add_vit_feature": True,
        "deform_num_heads": 16,
        "n_points": 4,
        "init_values": 0,
        "with_cffn": False,
        "cffn_ratio": 0.25,
        "deform_ratio": 0.5,
        "use_extra_extractor": True,
        "with_cp": False,
    },
    "prithvi_eo_v2_300_tl": {
        "interaction_indexes": [[0, 5], [6, 11], [12, 17], [18, 23]],
        "conv_inplane": 64,
        "add_vit_feature": True,
        "deform_num_heads": 16,
        "n_points": 4,
        "init_values": 0,
        "with_cffn": False,
        "cffn_ratio": 0.25,
        "deform_ratio": 0.5,
        "use_extra_extractor": True,
        "with_cp": False,
    },
}


def checkpoint_filter_fn_vit(
    state_dict, model: PrithviViT, pretrained_bands: list[HLSBands | int], model_bands: list[HLSBands | int]
) -> dict:
    """Encoder only model"""

    clean_dict = {}
    for k, v in state_dict.items():
        if "_timm_module." in k:  # Backwards compatibility for old model checkpoints
            k = k.replace("_timm_module.", "")

        if "pos_embed" in k:
            v = model.pos_embed  # pos_embed depends on num_frames and is fixed.
        if "decoder" in k or "_dec" in k or k == "mask_token":
            continue  # Drop decoder weights

        if not model.temporal_encoding and "temporal_embed" in k:
            continue
        if not model.location_encoding and "location_embed" in k:
            continue

        if k.startswith("encoder."):
            clean_dict[k.replace("encoder.", "")] = v  # Convert Prithvi MAE to Prithvi ViT
        else:
            clean_dict[k] = v
    
    for k, v in model.state_dict().items():
        if "vpt_prompt_embeddings" in k:
            clean_dict[k] = v

    state_dict = clean_dict

    state_dict = select_patch_embed_weights(state_dict, model, pretrained_bands, model_bands, encoder_only=True)

    return state_dict


def checkpoint_filter_fn_mae(
    state_dict, model: PrithviMAE, pretrained_bands: list[HLSBands | int], model_bands: list[HLSBands | int]
) -> dict:
    """Encoder-decoder model"""

    clean_dict = {}
    for k, v in state_dict.items():
        if "_timm_module." in k:  # Backwards compatibility for old model checkpoints
            k = k.replace("_timm_module.", "")

        # pos_embed depends on num_frames and is fixed.
        if "decoder_pos_embed" in k:
            v = model.decoder.decoder_pos_embed
        elif "pos_embed" in k:
            v = model.encoder.pos_embed

        if not model.encoder.temporal_encoding and "temporal_embed" in k:
            continue
        if not model.encoder.location_encoding and "location_embed" in k:
            continue

        if k.startswith("encoder.") or k.startswith("decoder."):
            clean_dict[k] = v  # Weights in Prithvi MAE format
        # Convert Prithvi V1 weights
        elif "decoder" in k or "_dec" in k or k == "mask_token":
            clean_dict["decoder." + k] = v
        else:
            clean_dict["encoder." + k] = v

    state_dict = clean_dict

    state_dict = select_patch_embed_weights(state_dict, model, pretrained_bands, model_bands, encoder_only=False)

    return state_dict


def checkpoint_filter_fn_vit_adapter(
    state_dict: dict,
    model: PrithviViTAdapter,
    pretrained_bands: list[HLSBands | str | int],
    model_bands: list[HLSBands | int],
) -> dict:
    state_dict = checkpoint_filter_fn_vit(state_dict, model, pretrained_bands, model_bands)
    clean_dict = {}
    for k, v in state_dict.items():
        if "cls_token" in k:
            continue
        if "norm." in k:
            continue
        clean_dict[k] = v

    for k, v in model.state_dict().items():
        if any(k.startswith(prefix) for prefix in PrithviViTAdapter.extra_layers):
            clean_dict[k] = v

    return clean_dict


def _create_prithvi(
    variant: str,
    pretrained: bool = False,  # noqa: FBT001, FBT002    
    model_bands: list[HLSBands | int] | None = None,
    ckpt_path: str = None,
    pretrained_bands: list[HLSBands | str | int] | None = None,
    num_frames: int = 1,
    encoder_only: bool = True,
    vit_adapter: bool = False,
    **kwargs,
) -> PrithviViT | PrithviMAE | PrithviViTAdapter:
    """
    Build PrithviViT and PrithviMAE models.
    By default, encoder_only is set to True and a ViT is returned.
    """

    # Load default config
    model_args = prithvi_cfgs[variant].copy()

    if vit_adapter:
        if variant not in prithvi_adapter_cfgs:
            msg = (
                f"ViT Adapter not available for variant {variant}. "
                f"Available variants: {list(prithvi_adapter_cfgs.keys())}."
            )
            raise ValueError(msg)
        model_args |= prithvi_adapter_cfgs[variant].copy()

    # Backwards compatibility from timm (pretrained_cfg_overlay={"file": "<path to weights>"}) TODO: Remove before v1.0
    if "pretrained_cfg_overlay" in kwargs:
        warnings.warn(
            "pretrained_cfg_overlay is deprecated and will be removed in a future version, "
            "use ckpt_path=<file path> instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        if ckpt_path is not None:
            warnings.warn("pretrained_cfg_overlay and ckpt_path are provided, ignoring pretrained_cfg_overlay.")
        elif "file" not in kwargs["pretrained_cfg_overlay"]:
            warnings.warn("pretrained_cfg_overlay does not include 'file path', ignoring pretrained_cfg_overlay.")
        else:
            ckpt_path = kwargs.pop("pretrained_cfg_overlay")["file"]

    pretrained_bands = pretrained_bands or model_args.get("bands", PRETRAINED_BANDS)

    if model_bands is None:
        model_bands: list[HLSBands | int] = pretrained_bands
        logger.info(
            f"model_bands not passed. Assuming bands are ordered in the same way as {pretrained_bands}."
            f"Pretrained patch_embed layer may be misaligned with current bands"
        )

    else:
        model_bands = [HLSBands.try_convert_to_hls_bands_enum(b) for b in model_bands]
        model_bands = generate_bands_intervals(model_bands)

    kwargs["in_chans"] = len(model_bands)
    kwargs["num_frames"] = num_frames
    model_args.update(kwargs)

    prithvi_model_class: type[nn.Module]
    checkpoint_filter_wrapper_fn: Callable
    if vit_adapter:
        if not encoder_only:
            msg = "PrithviViTAdapter only supports encoder_only=True"
            raise ValueError(msg)
        prithvi_model_class = PrithviViTAdapter
        checkpoint_filter_wrapper_fn = checkpoint_filter_fn_vit_adapter
    elif encoder_only:
        prithvi_model_class = PrithviViT
        checkpoint_filter_wrapper_fn = checkpoint_filter_fn_vit
    else:
        prithvi_model_class = PrithviMAE
        checkpoint_filter_wrapper_fn = checkpoint_filter_fn_mae

    model = prithvi_model_class(**model_args)

    if pretrained:
        if ckpt_path is not None:
            # Load model from checkpoint
            state_dict = torch.load(ckpt_path, map_location="cpu", weights_only=True)
            state_dict = checkpoint_filter_wrapper_fn(state_dict, model, pretrained_bands, model_bands)

            loaded_keys = model.load_state_dict(state_dict, strict=False)
            if loaded_keys.missing_keys:
                logger.warning(f"Missing keys in ckpt_path {ckpt_path}: {loaded_keys.missing_keys}")
            if loaded_keys.unexpected_keys:
                logger.warning(f"Unexpected keys in ckpt_path {ckpt_path}: {loaded_keys.unexpected_keys}")
        else:
            assert variant in pretrained_weights, (
                f"No pre-trained model found for variant {variant} (pretrained models: {pretrained_weights.keys()})"
            )

            try:
                # Download config.json to count model downloads
                _ = hf_hub_download(repo_id=pretrained_weights[variant]["hf_hub_id"], filename="config.json")
                # Load model from Hugging Face
                pretrained_path = hf_hub_download(
                    repo_id=pretrained_weights[variant]["hf_hub_id"],
                    filename=pretrained_weights[variant]["hf_hub_filename"],
                )
                state_dict = torch.load(pretrained_path, map_location="cpu", weights_only=True)
                state_dict = checkpoint_filter_wrapper_fn(state_dict, model, pretrained_bands, model_bands)
                model.load_state_dict(state_dict, strict=True)
            except RuntimeError as e:
                logger.error(f"Failed to load the pre-trained weights for {variant}.")
                raise e
    elif ckpt_path is not None:
        logger.warning(f"ckpt_path is provided but pretrained is set to False, ignoring ckpt_path {ckpt_path}.")

    # TODO Renanme to model.bands? 
    model.model_bands = model_bands
    model.pretrained_bands = pretrained_bands

    assert encoder_only or "out_indices" not in kwargs, "out_indices provided for a MAE model."
    if vit_adapter:
        if "out_indices" in kwargs:
            msg = "out_indices should not be provided for ViTAdapter"
            raise ValueError(msg)
        model.forward = model.forward_features
        model.out_indices = [indexes[-1] for indexes in model.interaction_indexes]
    elif encoder_only:
        default_out_indices = list(range(len(model.blocks)))
        out_indices = kwargs.pop("out_indices", default_out_indices)

        def forward_filter_indices(*args, **kwargs):
            features = model.forward_features(*args, **kwargs)
            return [features[i] for i in out_indices]

        model.forward = forward_filter_indices
        model.out_indices = out_indices

    return model


@TERRATORCH_BACKBONE_REGISTRY.register
def prithvi_eo_tiny(
    pretrained: bool = False,  # noqa: FBT001, FBT002
    bands: list[HLSBands] | None = None,
    vit_adapter: bool = False,
    **kwargs,
) -> PrithviViT | PrithviViTAdapter:
    return _create_prithvi(
        "prithvi_eo_tiny", pretrained=pretrained, vit_adapter=vit_adapter, **dict({"model_bands": bands}), **kwargs
    )


@TERRATORCH_BACKBONE_REGISTRY.register
def prithvi_eo_v1_100(
    pretrained: bool = False,  # noqa: FBT001, FBT002
    bands: list[HLSBands] | None = None,
    vit_adapter: bool = False,
    **kwargs,
) -> PrithviViT | PrithviViTAdapter:
    return _create_prithvi(
        "prithvi_eo_v1_100", pretrained=pretrained, vit_adapter=vit_adapter, **dict({"model_bands": bands}), **kwargs
    )



@TERRATORCH_BACKBONE_REGISTRY.register
def prithvi_eo_v2_300(
    pretrained: bool = False,  # noqa: FBT001, FBT002
    bands: list[HLSBands] | None = None,
    vit_adapter: bool = False,
    **kwargs,
) -> PrithviViT | PrithviViTAdapter:
    return _create_prithvi(
        "prithvi_eo_v2_300", pretrained=pretrained, vit_adapter=vit_adapter, **dict({"model_bands": bands}), **kwargs
    )


@TERRATORCH_BACKBONE_REGISTRY.register
def prithvi_eo_v2_600(
    pretrained: bool = False,  # noqa: FBT001, FBT002
    bands: list[HLSBands] | None = None,
    vit_adapter: bool = False,
    **kwargs,
) -> PrithviViT | PrithviViTAdapter:
    return _create_prithvi(
        "prithvi_eo_v2_600", pretrained=pretrained, vit_adapter=vit_adapter, **dict({"model_bands": bands}), **kwargs
    )


@TERRATORCH_BACKBONE_REGISTRY.register
def prithvi_eo_v2_300_tl(
    pretrained: bool = False,  # noqa: FBT001, FBT002
    bands: list[HLSBands] | None = None,
    vit_adapter: bool = False,
    **kwargs,
) -> PrithviViT | PrithviViTAdapter:
    return _create_prithvi(
        "prithvi_eo_v2_300_tl", pretrained=pretrained, vit_adapter=vit_adapter, **dict({"model_bands": bands}), **kwargs
    )

  
@TERRATORCH_BACKBONE_REGISTRY.register
def prithvi_eo_v2_600_tl(
    pretrained: bool = False,  # noqa: FBT001, FBT002
    bands: list[HLSBands] | None = None,
    vit_adapter: bool = False,
    **kwargs,
) -> PrithviViT | PrithviViTAdapter:
    return _create_prithvi(
        "prithvi_eo_v2_600_tl", pretrained=pretrained, vit_adapter=vit_adapter, **dict({"model_bands": bands}), **kwargs
    )


@ TERRATORCH_FULL_MODEL_REGISTRY.register
def prithvi_eo_v1_100_mae(
    pretrained: bool = False,  # noqa: FBT001, FBT002
    bands: list[HLSBands] | None = None,
    encoder_only: bool = False,
    **kwargs,
) -> PrithviMAE:
    if encoder_only:
        raise ValueError("Please use 'prithvi_eo_v1_100' for encoder only models.")
    model_bands = kwargs.get('model_bands', bands)

    return _create_prithvi("prithvi_eo_v1_100", pretrained=pretrained, bands=bands, encoder_only=encoder_only,
                           **kwargs)


@TERRATORCH_FULL_MODEL_REGISTRY.register
def prithvi_eo_v1_100_mae(
        pretrained: bool = False,  # noqa: FBT001, FBT002
        bands: list[HLSBands] | None = None,
        encoder_only: bool = False,
        **kwargs,
) -> PrithviMAE:
    if encoder_only:
        raise ValueError("Please use 'prithvi_eo_v1_100' for encoder only models.")
    model_bands = kwargs.get('model_bands', bands)

    return _create_prithvi("prithvi_eo_v1_100", pretrained=pretrained, model_bands=model_bands, encoder_only=encoder_only,
                           **kwargs)


@TERRATORCH_FULL_MODEL_REGISTRY.register
def prithvi_eo_v2_300_mae(
        pretrained: bool = False,
        bands: list[HLSBands] | None = None,
        encoder_only: bool = False,
        **kwargs,
) -> PrithviMAE:
    if encoder_only:
        raise ValueError("Please use 'prithvi_eo_v2_300' for encoder only models.")        
    model_bands = kwargs.get('model_bands', bands)

    return _create_prithvi("prithvi_eo_v2_300", pretrained=pretrained, model_bands=model_bands, encoder_only=encoder_only,
                           **kwargs)


@TERRATORCH_FULL_MODEL_REGISTRY.register
def prithvi_eo_v2_300_tl_mae(
        pretrained: bool = False,
        bands: list[HLSBands] | None = None,
        encoder_only: bool = False,
        **kwargs,
) -> PrithviMAE:
    if encoder_only:
        raise ValueError("Please use 'prithvi_eo_v2_300_tl' for encoder only models.")
    model_bands = kwargs.get('model_bands', bands)

    return _create_prithvi("prithvi_eo_v2_300_tl", pretrained=pretrained, model_bands=model_bands, encoder_only=encoder_only,
                           **kwargs)


@TERRATORCH_FULL_MODEL_REGISTRY.register
def prithvi_eo_v2_600_mae(
        pretrained: bool = False,
        bands: list[HLSBands] | None = None,
        encoder_only: bool = False,
        **kwargs,
) -> PrithviMAE:
    if encoder_only:
        raise ValueError("Please use 'prithvi_eo_v2_600' for encoder only models.")
    model_bands = kwargs.get('model_bands', bands)

    return _create_prithvi("prithvi_eo_v2_600", pretrained=pretrained, model_bands=model_bands, encoder_only=encoder_only,
                           **kwargs)

@TERRATORCH_FULL_MODEL_REGISTRY.register
def prithvi_eo_v2_600_tl_mae(
        pretrained: bool = False,
        bands: list[HLSBands] | None = None,
        encoder_only: bool = False,
        **kwargs,
) -> PrithviMAE:
    if encoder_only:
        raise ValueError("Please use 'prithvi_eo_v2_600_tl' for encoder only models.")
    model_bands = kwargs.get('model_bands', bands)

    return _create_prithvi("prithvi_eo_v2_600_tl", pretrained=pretrained, model_bands=model_bands, encoder_only=encoder_only,
                           **kwargs)


# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.

"""Module containing generic dataset classes"""

import glob
import os
from abc import ABC
from pathlib import Path
from typing import Any

import albumentations as A
import matplotlib as mpl
import numpy as np
import rioxarray
import xarray as xr
from einops import rearrange
from matplotlib import pyplot as plt
from matplotlib.figure import Figure
from matplotlib.patches import Rectangle
from torch import Tensor
from torchgeo.datasets import NonGeoDataset

from terratorch.datasets.utils import HLSBands, default_transform, filter_valid_files, generate_bands_intervals


class GenericPixelWiseDataset(NonGeoDataset, ABC):
    """
    This is a generic dataset class to be used for instantiating datasets from arguments.
    Ideally, one would create a dataset class specific to a dataset.
    """

    def __init__(
        self,
        data_root: Path,
        label_data_root: Path | None = None,
        image_grep: str | None = "*",
        label_grep: str | None = "*",
        split: Path | None = None,
        ignore_split_file_extensions: bool = True,
        allow_substring_split_file: bool = True,
        rgb_indices: list[int] | None = None,
        dataset_bands: list[HLSBands | int | tuple[int, int] | str] | None = None,
        output_bands: list[HLSBands | int | tuple[int, int] | str] | None = None,
        constant_scale: float = 1,
        transform: A.Compose | None = None,
        no_data_replace: float | None = None,
        no_label_replace: int | None = None,
        expand_temporal_dimension: bool = False,
        reduce_zero_label: bool = False,
    ) -> None:
        """Constructor

        Args:
            data_root (Path): Path to data root directory
            label_data_root (Path, optional): Path to data root directory with labels.
                If not specified, will use the same as for images.
            image_grep (str, optional): Regular expression appended to data_root to find input images.
                Defaults to "*".
            label_grep (str, optional): Regular expression appended to data_root to find ground truth masks.
                Defaults to "*".
            split (Path, optional): Path to file containing files to be used for this split.
                The file should be a new-line separated prefixes contained in the desired files.
                Files will be seached using glob with the form Path(data_root).glob(prefix + [image or label grep])
            ignore_split_file_extensions (bool, optional): Whether to disregard extensions when using the split
                file to determine which files to include in the dataset.
                E.g. necessary for Eurosat, since the split files specify ".jpg" but files are
                actually ".jpg". Defaults to True.
            allow_substring_split_file (bool, optional): Whether the split files contain substrings
                that must be present in file names to be included (as in mmsegmentation), or exact
                matches (e.g. eurosat). Defaults to True.
            rgb_indices (list[str], optional): Indices of RGB channels. Defaults to [0, 1, 2].
            dataset_bands (list[HLSBands | int | tuple[int, int] | str] | None): Bands present in the dataset. This parameter names input channels (bands) using HLSBands, ints, int ranges, or strings, so that they can then be refered to by output_bands. Defaults to None.
            output_bands (list[HLSBands | int | tuple[int, int] | str] | None): Bands that should be output by the dataset as named by dataset_bands.
            constant_scale (float): Factor to multiply image values by. Defaults to 1.
            transform (Albumentations.Compose | None): Albumentations transform to be applied.
                Should end with ToTensorV2(). If used through the generic_data_module,
                should not include normalization. Not supported for multi-temporal data.
                Defaults to None, which simply applies ToTensorV2().
            no_data_replace (float | None): Replace nan values in input images with this value. If none, does no replacement. Defaults to None.
            no_label_replace (int | None): Replace nan values in label with this value. If none, does no replacement. Defaults to -1.
            expand_temporal_dimension (bool): Go from shape (time*channels, h, w) to (channels, time, h, w).
                Defaults to False.
            reduce_zero_label (bool): Subtract 1 from all labels. Useful when labels start from 1 instead of the
                expected 0. Defaults to False.
        """
        super().__init__()

        self.split_file = split

        label_data_root = label_data_root if label_data_root is not None else data_root
        self.image_files = sorted(glob.glob(os.path.join(data_root, image_grep)))
        self.constant_scale = constant_scale
        self.no_data_replace = no_data_replace
        self.no_label_replace = no_label_replace
        self.segmentation_mask_files = sorted(glob.glob(os.path.join(label_data_root, label_grep)))
        self.reduce_zero_label = reduce_zero_label
        self.expand_temporal_dimension = expand_temporal_dimension

        if self.expand_temporal_dimension and output_bands is None:
            msg = "Please provide output_bands when expand_temporal_dimension is True"
            raise Exception(msg)
        if self.split_file is not None:
            with open(self.split_file) as f:
                split = f.readlines()
            valid_files = {rf"{substring.strip()}" for substring in split}
            self.image_files = filter_valid_files(
                self.image_files,
                valid_files=valid_files,
                ignore_extensions=ignore_split_file_extensions,
                allow_substring=allow_substring_split_file,
            )
            self.segmentation_mask_files = filter_valid_files(
                self.segmentation_mask_files,
                valid_files=valid_files,
                ignore_extensions=ignore_split_file_extensions,
                allow_substring=allow_substring_split_file,
            )

        # We don't define a split file for prediction
        if not self.split_file:
            # When prediction is enabled, we don't have mask files, so
            # we need to provide a way to run the dataloder in these cases.
            if not self.segmentation_mask_files:
                self.segmentation_mask_files = self.image_files
                # The masks can be `None` since they won't be used in fact. 

        self.rgb_indices = [0, 1, 2] if rgb_indices is None else rgb_indices

        self.dataset_bands = generate_bands_intervals(dataset_bands)
        self.output_bands = generate_bands_intervals(output_bands)

        if self.output_bands and not self.dataset_bands:
            msg = "If output bands provided, dataset_bands must also be provided"
            return Exception(msg)  # noqa: PLE0101

        # There is a special condition if the bands are defined as simple strings.
        if self.output_bands:
            if len(set(self.output_bands) & set(self.dataset_bands)) != len(self.output_bands):
                msg = "Output bands must be a subset of dataset bands"
                raise Exception(msg)

            self.filter_indices = [self.dataset_bands.index(band) for band in self.output_bands]

        else:
            self.filter_indices = None

        # If no transform is given, apply only to transform to torch tensor
        self.transform = transform if transform else default_transform
        # self.transform = transform if transform else ToTensorV2()

        import warnings

        import rasterio

        warnings.filterwarnings("ignore", category=rasterio.errors.NotGeoreferencedWarning)

    def __len__(self) -> int:
        return len(self.image_files)

    def __getitem__(self, index: int) -> dict[str, Any]:
        image = self._load_file(self.image_files[index], nan_replace=self.no_data_replace).to_numpy()
        # to channels last
        if self.expand_temporal_dimension:
            image = rearrange(image, "(channels time) h w -> channels time h w", channels=len(self.output_bands))
        image = np.moveaxis(image, 0, -1)

        if self.filter_indices:
            image = image[..., self.filter_indices]
        output = {
            "image": image.astype(np.float32) * self.constant_scale,
        }
        if self.segmentation_mask_files:
            mask = self._load_file(self.segmentation_mask_files[index], nan_replace=self.no_label_replace)
            output["mask"] = mask.to_numpy()[0]
            if self.reduce_zero_label:
                output["mask"] -= 1
        if self.transform:
            output = self.transform(**output)
        output["filename"] = self.image_files[index]

        return output

    def _load_file(self, path, nan_replace: int | float | None = None) -> xr.DataArray:
        data = rioxarray.open_rasterio(path, masked=True)
        if nan_replace is not None:
            data = data.fillna(nan_replace)
        return data


class GenericNonGeoSegmentationDataset(GenericPixelWiseDataset):
    """GenericNonGeoSegmentationDataset"""

    def __init__(
        self,
        data_root: Path,
        num_classes: int,
        label_data_root: Path | None = None,
        image_grep: str | None = "*",
        label_grep: str | None = "*",
        split: Path | None = None,
        ignore_split_file_extensions: bool = True,
        allow_substring_split_file: bool = True,
        rgb_indices: list[str] | None = None,
        dataset_bands: list[HLSBands | int | tuple[int, int] | str] | None = None,
        output_bands: list[HLSBands | int | tuple[int, int] | str] | None = None,
        class_names: list[str] | None = None,
        constant_scale: float = 1,
        transform: A.Compose | None = None,
        no_data_replace: float | None = None,
        no_label_replace: int | None = None,
        expand_temporal_dimension: bool = False,
        reduce_zero_label: bool = False,
    ) -> None:
        """Constructor

        Args:
            data_root (Path): Path to data root directory
            num_classes (int): Number of classes in the dataset
            label_data_root (Path, optional): Path to data root directory with labels.
                If not specified, will use the same as for images.
            image_grep (str, optional): Regular expression appended to data_root to find input images.
                Defaults to "*".
            label_grep (str, optional): Regular expression appended to data_root to find ground truth masks.
                Defaults to "*".
            split (Path, optional): Path to file containing files to be used for this split.
                The file should be a new-line separated prefixes contained in the desired files.
                Files will be seached using glob with the form Path(data_root).glob(prefix + [image or label grep])
            ignore_split_file_extensions (bool, optional): Whether to disregard extensions when using the split
                file to determine which files to include in the dataset.
                E.g. necessary for Eurosat, since the split files specify ".jpg" but files are
                actually ".jpg". Defaults to True
            allow_substring_split_file (bool, optional): Whether the split files contain substrings
                that must be present in file names to be included (as in mmsegmentation), or exact
                matches (e.g. eurosat). Defaults to True.
            rgb_indices (list[str], optional): Indices of RGB channels. Defaults to [0, 1, 2].
            dataset_bands (list[HLSBands | int] | None): Bands present in the dataset.
            output_bands (list[HLSBands | int] | None): Bands that should be output by the dataset.
            class_names (list[str], optional): Class names. Defaults to None.
            constant_scale (float): Factor to multiply image values by. Defaults to 1.
            transform (Albumentations.Compose | None): Albumentations transform to be applied.
                Should end with ToTensorV2(). If used through the generic_data_module,
                should not include normalization. Not supported for multi-temporal data.
                Defaults to None, which simply applies ToTensorV2().
            no_data_replace (float | None): Replace nan values in input images with this value. If none, does no replacement. Defaults to None.
            no_label_replace (int | None): Replace nan values in label with this value. If none, does no replacement. Defaults to None.
            expand_temporal_dimension (bool): Go from shape (time*channels, h, w) to (channels, time, h, w).
                Defaults to False.
            reduce_zero_label (bool): Subtract 1 from all labels. Useful when labels start from 1 instead of the
                expected 0. Defaults to False.
        """
        super().__init__(
            data_root,
            label_data_root=label_data_root,
            image_grep=image_grep,
            label_grep=label_grep,
            split=split,
            ignore_split_file_extensions=ignore_split_file_extensions,
            allow_substring_split_file=allow_substring_split_file,
            rgb_indices=rgb_indices,
            dataset_bands=dataset_bands,
            output_bands=output_bands,
            constant_scale=constant_scale,
            transform=transform,
            no_data_replace=no_data_replace,
            no_label_replace=no_label_replace,
            expand_temporal_dimension=expand_temporal_dimension,
            reduce_zero_label=reduce_zero_label,
        )
        self.num_classes = num_classes
        self.class_names = class_names

    def __getitem__(self, index: int) -> dict[str, Any]:
        item = super().__getitem__(index)
        if "mask" in item:
            item["mask"] = item["mask"].long()
        return item

    def plot(self, sample: dict[str, Tensor], suptitle: str | None = None, show_axes: bool | None = False) -> Figure:
        """Plot a sample from the dataset.

        Args:
            sample: a sample returned by :meth:`__getitem__`
            suptitle: optional string to use as a suptitle
            show_axes: whether to show axes or not

        Returns:
            a matplotlib Figure with the rendered sample

        .. versionadded:: 0.2
        """
        image = sample["image"]
        if len(image.shape) == 5:
            return
        if isinstance(image, Tensor):
            image = image.numpy()
        image = image.take(self.rgb_indices, axis=0)
        image = np.transpose(image, (1, 2, 0))
        image = (image - image.min(axis=(0, 1))) * (1 / image.max(axis=(0, 1)))
        image = np.clip(image, 0, 1)

        label_mask = sample["mask"]
        if isinstance(label_mask, Tensor):
            label_mask = label_mask.numpy()

        showing_predictions = "prediction" in sample
        if showing_predictions:
            prediction_mask = sample["prediction"]
            if isinstance(prediction_mask, Tensor):
                prediction_mask = prediction_mask.numpy()

        return self._plot_sample(
            image,
            label_mask,
            self.num_classes,
            prediction=prediction_mask if showing_predictions else None,
            suptitle=suptitle,
            class_names=self.class_names,
            show_axes=show_axes,
        )

    @staticmethod
    def _plot_sample(image, label, num_classes, prediction=None, suptitle=None, class_names=None, show_axes=False):
        num_images = 5 if prediction is not None else 4
        fig, ax = plt.subplots(1, num_images, figsize=(12, 10), layout="compressed")
        axes_visibility = "on" if show_axes else "off"

        # for legend
        ax[0].axis("off")

        norm = mpl.colors.Normalize(vmin=0, vmax=num_classes - 1)
        ax[1].axis(axes_visibility)
        ax[1].title.set_text("Image")
        ax[1].imshow(image)

        ax[2].axis(axes_visibility)
        ax[2].title.set_text("Ground Truth Mask")
        ax[2].imshow(label, cmap="jet", norm=norm)

        ax[3].axis(axes_visibility)
        ax[3].title.set_text("GT Mask on Image")
        ax[3].imshow(image)
        ax[3].imshow(label, cmap="jet", alpha=0.3, norm=norm)

        if prediction is not None:
            ax[4].axis(axes_visibility)
            ax[4].title.set_text("Predicted Mask")
            ax[4].imshow(prediction, cmap="jet", norm=norm)

        cmap = plt.get_cmap("jet")
        legend_data = []
        for i, _ in enumerate(range(num_classes)):
            class_name = class_names[i] if class_names else str(i)
            data = [i, cmap(norm(i)), class_name]
            legend_data.append(data)
        handles = [Rectangle((0, 0), 1, 1, color=tuple(v for v in c)) for k, c, n in legend_data]
        labels = [n for k, c, n in legend_data]
        ax[0].legend(handles, labels, loc="center")
        if suptitle is not None:
            plt.suptitle(suptitle)
        return fig


class GenericNonGeoPixelwiseRegressionDataset(GenericPixelWiseDataset):
    """GenericNonGeoPixelwiseRegressionDataset"""

    def __init__(
        self,
        data_root: Path,
        label_data_root: Path | None = None,
        image_grep: str | None = "*",
        label_grep: str | None = "*",
        split: Path | None = None,
        ignore_split_file_extensions: bool = True,
        allow_substring_split_file: bool = True,
        rgb_indices: list[int] | None = None,
        dataset_bands: list[HLSBands | int | tuple[int, int] | str] | None = None,
        output_bands: list[HLSBands | int | tuple[int, int] | str] | None = None,
        constant_scale: float = 1,
        transform: A.Compose | None = None,
        no_data_replace: float | None = None,
        no_label_replace: int | None = None,
        expand_temporal_dimension: bool = False,
        reduce_zero_label: bool = False,
    ) -> None:
        """Constructor

        Args:
            data_root (Path): Path to data root directory
            label_data_root (Path, optional): Path to data root directory with labels.
                If not specified, will use the same as for images.
            image_grep (str, optional): Regular expression appended to data_root to find input images.
                Defaults to "*".
            label_grep (str, optional): Regular expression appended to data_root to find ground truth masks.
                Defaults to "*".
            split (Path, optional): Path to file containing files to be used for this split.
                The file should be a new-line separated prefixes contained in the desired files.
                Files will be seached using glob with the form Path(data_root).glob(prefix + [image or label grep])
            ignore_split_file_extensions (bool, optional): Whether to disregard extensions when using the split
                file to determine which files to include in the dataset.
                E.g. necessary for Eurosat, since the split files specify ".jpg" but files are
                actually ".jpg". Defaults to True.
            allow_substring_split_file (bool, optional): Whether the split files contain substrings
                that must be present in file names to be included (as in mmsegmentation), or exact
                matches (e.g. eurosat). Defaults to True.
            rgb_indices (list[str], optional): Indices of RGB channels. Defaults to [0, 1, 2].
            dataset_bands (list[HLSBands | int] | None): Bands present in the dataset.
            output_bands (list[HLSBands | int] | None): Bands that should be output by the dataset.
            constant_scale (float): Factor to multiply image values by. Defaults to 1.
            transform (Albumentations.Compose | None): Albumentations transform to be applied.
                Should end with ToTensorV2(). If used through the generic_data_module,
                should not include normalization. Not supported for multi-temporal data.
                Defaults to None, which simply applies ToTensorV2().
            no_data_replace (float | None): Replace nan values in input images with this value. If none, does no replacement. Defaults to None.
            no_label_replace (int | None): Replace nan values in label with this value. If none, does no replacement. Defaults to None.
            expand_temporal_dimension (bool): Go from shape (time*channels, h, w) to (channels, time, h, w).
                Defaults to False.
            reduce_zero_label (bool): Subtract 1 from all labels. Useful when labels start from 1 instead of the
                expected 0. Defaults to False.
        """
        super().__init__(
            data_root,
            label_data_root=label_data_root,
            image_grep=image_grep,
            label_grep=label_grep,
            split=split,
            ignore_split_file_extensions=ignore_split_file_extensions,
            allow_substring_split_file=allow_substring_split_file,
            rgb_indices=rgb_indices,
            dataset_bands=dataset_bands,
            output_bands=output_bands,
            constant_scale=constant_scale,
            transform=transform,
            no_data_replace=no_data_replace,
            no_label_replace=no_label_replace,
            expand_temporal_dimension=expand_temporal_dimension,
            reduce_zero_label=reduce_zero_label,
        )

    def __getitem__(self, index: int) -> dict[str, Any]:
        item = super().__getitem__(index)
        if "mask" in item:
            item["mask"] = item["mask"].float()
        return item

    def plot(self, sample: dict[str, Tensor], suptitle: str | None = None, show_axes: bool | None = False) -> Figure:
        """Plot a sample from the dataset.

        Args:
            sample (dict[str, Tensor]): a sample returned by :meth:`__getitem__`
            suptitle (str|None): optional string to use as a suptitle
            show_axes (bool|None): whether to show axes or not

        Returns:
            a matplotlib Figure with the rendered sample

        .. versionadded:: 0.2
        """
        image = sample["image"]
        if len(image.shape) == 5:
            return
        if isinstance(image, Tensor):
            image = image.numpy()
        image = image.take(self.rgb_indices, axis=0)
        image = np.transpose(image, (1, 2, 0))
        image = (image - image.min(axis=(0, 1))) * (1 / image.max(axis=(0, 1)))
        image = np.clip(image, 0, 1)

        label_mask = sample["mask"]
        if isinstance(label_mask, Tensor):
            label_mask = label_mask.numpy()

        showing_predictions = "prediction" in sample
        if showing_predictions:
            prediction_mask = sample["prediction"]
            if isinstance(prediction_mask, Tensor):
                prediction_mask = prediction_mask.numpy()

        return self._plot_sample(
            image,
            label_mask,
            prediction=prediction_mask if showing_predictions else None,
            suptitle=suptitle,
            show_axes=show_axes,
        )

    @staticmethod
    def _plot_sample(image, label, prediction=None, suptitle=None, show_axes=False):
        num_images = 4 if prediction is not None else 3
        fig, ax = plt.subplots(1, num_images, figsize=(12, 10), layout="compressed")
        axes_visibility = "on" if show_axes else "off"

        norm = mpl.colors.Normalize(vmin=label.min(), vmax=label.max())
        ax[0].axis(axes_visibility)
        ax[0].title.set_text("Image")
        ax[0].imshow(image)

        ax[1].axis(axes_visibility)
        ax[1].title.set_text("Ground Truth Mask")
        ax[1].imshow(label, cmap="Greens", norm=norm)

        ax[2].axis(axes_visibility)
        ax[2].title.set_text("GT Mask on Image")
        ax[2].imshow(image)
        ax[2].imshow(label, cmap="Greens", alpha=0.3, norm=norm)
        # ax[2].legend()

        if prediction is not None:
            ax[3].axis(axes_visibility)
            ax[3].title.set_text("Predicted Mask")
            ax[3].imshow(prediction, cmap="Greens", norm=norm)

        if suptitle is not None:
            plt.suptitle(suptitle)
        return fig



import binascii
import json
import pickle

import albumentations as A
import h5py
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pytest
import rasterio
import torch
from albumentations.pytorch import ToTensorV2
from rasterio.transform import from_origin
from xarray import DataArray

from terratorch.datasets import (
    FireScarsNonGeo,
    MBeninSmallHolderCashewsNonGeo,
    MBigEarthNonGeo,
    MBrickKilnNonGeo,
    MChesapeakeLandcoverNonGeo,
    MEuroSATNonGeo,
    MForestNetNonGeo,
    MNeonTreeNonGeo,
    MNzCattleNonGeo,
    MPv4gerNonGeo,
    MPv4gerSegNonGeo,
    MSACropTypeNonGeo,
    MSo2SatNonGeo,
    MultiTemporalCropClassification,
    #Sen1Floods11NonGeo,
)
from terratorch.datasets.sen1floods11 import Sen1Floods11NonGeo

from terratorch.datasets.transforms import FlattenTemporalIntoChannels, UnflattenTemporalFromChannels


def create_dummy_tiff(path, width=100, height=100, count=6, dtype="uint8"):
    data = np.random.randint(0, 255, (count, height, width), dtype=dtype)
    transform = from_origin(0, 0, 1, 1)
    with rasterio.open(
        path, "w",
        driver="GTiff",
        height=height,
        width=width,
        count=count,
        dtype=dtype,
        crs="EPSG:4326",
        transform=transform
    ) as dst:
        dst.write(data)

@pytest.fixture(scope="function")
def mpv4ger_data_root(tmp_path):
    data_root = tmp_path / "m_pv4ger"
    data_directory = data_root / "m-pv4ger"
    data_directory.mkdir(parents=True, exist_ok=True)

    partition = {
        "train": ["10.0,20.0", "11.0,21.0"],
        "val": [],
        "test": []
    }
    partition_file = data_directory / "default_partition.json"
    with open(partition_file, "w") as f:
        json.dump(partition, f)

    bands = ["BLUE", "GREEN", "RED"]

    for img_id in partition["train"]:
        file_path = data_directory / f"{img_id}.hdf5"
        with h5py.File(file_path, "w") as h5file:

            for band in bands:
                h5file.create_dataset(band, data=np.random.rand(100, 100).astype(np.float32))


            attr_dict = {"label": np.random.randint(0, 10)}
            serialized_attr = pickle.dumps(attr_dict)
            hex_attr = binascii.hexlify(serialized_attr).decode("ascii")
            h5file.attrs["pickle"] = hex_attr

    return str(data_root)

@pytest.fixture(scope="function")
def fire_scars_data_root(tmp_path):
    data_root = tmp_path / "fire_scars"
    split = "train"
    split_dir = data_root / FireScarsNonGeo.splits[split]
    split_dir.mkdir(parents=True, exist_ok=True)

    for i in range(5):
        random_seq = "".join(np.random.choice(list("ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"), 5))
        year = 2021
        julian_day = i + 1
        date = f"{year}{julian_day:03d}"
        image_filename = f"subsetted_512x512_HLS.S30.T{random_seq}.{date}.v1.4_merged.tif"
        mask_filename = f"subsetted_512x512_HLS.S30.T{random_seq}.{date}.v1.4.mask.tif"
        image_path = split_dir / image_filename
        mask_path = split_dir / mask_filename

        create_dummy_tiff(image_path)
        create_dummy_tiff(mask_path, count=1, dtype="uint8")

    image_files = list(split_dir.glob("*_merged.tif"))
    mask_files = list(split_dir.glob("*.mask.tif"))

    assert len(image_files) == 5, f"Expected 5 image files, but found {len(image_files)}"
    assert len(mask_files) == 5, f"Expected 5 mask files, but found {len(mask_files)}"

    return str(data_root)

@pytest.fixture(scope="function")
def m_bigearth_data_root(tmp_path):
    data_root = tmp_path / "m_bigearthnet"
    data_directory = data_root / MBigEarthNonGeo.data_dir
    data_directory.mkdir(parents=True, exist_ok=True)

    label_map = {f"image_{i}": [0, 1] for i in range(2)}
    label_map_path = data_directory / MBigEarthNonGeo.label_map_file
    with open(label_map_path, "w") as f:
        json.dump(label_map, f)

    partition = {"train": ["image_0", "image_1"], "val": [], "test": []}
    partition_file = data_directory / MBigEarthNonGeo.partition_file_template.format(partition="default")
    with open(partition_file, "w") as f:
        json.dump(partition, f)

    bands = [
        "COASTAL_AEROSOL", "BLUE", "GREEN", "RED", "RED_EDGE_1",
        "RED_EDGE_2", "RED_EDGE_3", "NIR_BROAD", "NIR_NARROW",
        "WATER_VAPOR", "SWIR_1", "SWIR_2", "CLOUD_PROBABILITY"
    ]

    for img in ["image_0", "image_1"]:
        file_path = data_directory / f"{img}.hdf5"
        with h5py.File(file_path, "w") as h5file:
            for band in bands:
                h5file.create_dataset(band, data=np.random.rand(100, 100).astype(np.float32))

    image_files = list(data_directory.glob("*.hdf5"))
    assert len(image_files) == 2, f"Expected 2 image files, but found {len(image_files)}"

    return str(data_root)

@pytest.fixture(scope="function")
def m_forestnet_data_root(tmp_path):
    data_root = tmp_path / "m_forestnet"
    data_directory = data_root / MForestNetNonGeo.data_dir
    data_directory.mkdir(parents=True, exist_ok=True)

    partition = {
        "train": ["image_0", "image_1"],
        "val": [],
        "test": []
    }
    partition_file = data_directory / MForestNetNonGeo.partition_file_template.format(partition="default")
    with open(partition_file, "w") as f:
        json.dump(partition, f)

    bands = ["BLUE", "GREEN", "RED", "NIR", "SWIR_1", "SWIR_2"]
    label_map = {"image_0": 0, "image_1": 1}

    for img in ["image_0", "image_1"]:
        file_path = data_directory / f"{img}.hdf5"
        with h5py.File(file_path, "w") as h5file:
            for band in bands:
                h5file.create_dataset(band, data=np.random.rand(100, 100).astype(np.float32))

            label_dict = {"label": label_map[img]}
            pickle_bytes = pickle.dumps(label_dict)
            pickle_str = repr(pickle_bytes)
            h5file.attrs["pickle"] = pickle_str

    image_files = list(data_directory.glob("*.hdf5"))
    assert len(image_files) == 2, f"Expected 2 image files, but found {len(image_files)}"

    return str(data_root)

@pytest.fixture(scope="function")
def mnz_cattle_data_root(tmp_path):
    data_root = tmp_path / "m_nz_cattle"
    data_directory = data_root / MNzCattleNonGeo.data_dir
    data_directory.mkdir(parents=True, exist_ok=True)

    partition = {
        "train": ["image_0", "image_1"],
        "val": [],
        "test": []
    }
    partition_file = data_directory / MNzCattleNonGeo.partition_file_template.format(partition="default")
    with open(partition_file, "w") as f:
        json.dump(partition, f)

    bands = ["BLUE_2023-01-01", "GREEN_2023-01-01", "RED_2023-01-01"]

    for img in ["image_0", "image_1"]:
        file_path = data_directory / f"{img}.hdf5"
        with h5py.File(file_path, "w") as h5file:
            for band in bands:
                h5file.create_dataset(band, data=np.random.rand(100, 100).astype(np.float32))

            mask = np.random.randint(0, 2, size=(100, 100)).astype(np.uint8)
            h5file.create_dataset("label", data=mask)

    image_files = list(data_directory.glob("*.hdf5"))
    assert len(image_files) == 2, f"Expected 2 image files, but found {len(image_files)}"

    return str(data_root)

@pytest.fixture(scope="function")
def brickkiln_data_root(tmp_path):
    data_root = tmp_path / "m_brick_kiln"
    data_directory = data_root / MBrickKilnNonGeo.data_dir
    data_directory.mkdir(parents=True, exist_ok=True)

    partition = {
        "train": ["image_0", "image_1"],
        "val": [],
        "test": []
    }
    partition_file = data_directory / MBrickKilnNonGeo.partition_file_template.format(partition="default")
    with open(partition_file, "w") as f:
        json.dump(partition, f)

    bands = [
        "COASTAL_AEROSOL", "BLUE", "GREEN", "RED", "RED_EDGE_1", "RED_EDGE_2", "RED_EDGE_3",
        "NIR_BROAD", "NIR_NARROW", "WATER_VAPOR", "CIRRUS", "SWIR_1", "SWIR_2"
    ]

    for img in ["image_0", "image_1"]:
        file_path = data_directory / f"{img}.hdf5"
        with h5py.File(file_path, "w") as h5file:
            for band in bands:
                h5file.create_dataset(band, data=np.random.rand(100, 100).astype(np.float32))

            attr_dict = {"label": 1 if img == "image_0" else 0}
            h5file.attrs["pickle"] = str(pickle.dumps(attr_dict))

    return str(data_root)

@pytest.fixture(scope="function")
def neontree_data_root(tmp_path):
    data_root = tmp_path / "m_neon_tree"
    data_directory = data_root / MNeonTreeNonGeo.data_dir
    data_directory.mkdir(parents=True, exist_ok=True)

    partition = {
        "train": ["image_0", "image_1"],
        "val": [],
        "test": []
    }
    partition_file = data_directory / MNeonTreeNonGeo.partition_file_template.format(partition="default")
    with open(partition_file, "w") as f:
        json.dump(partition, f)

    bands = ["BLUE", "CANOPY_HEIGHT_MODEL", "GREEN", "NEON", "RED"]

    for img in ["image_0", "image_1"]:
        file_path = data_directory / f"{img}.hdf5"
        with h5py.File(file_path, "w") as h5file:
            for band in bands:
                h5file.create_dataset(band, data=np.random.rand(100, 100).astype(np.float32))

            mask = np.random.randint(0, 3, size=(100, 100), dtype=np.uint8)
            h5file.create_dataset("label", data=mask)

    return str(data_root)

@pytest.fixture(scope="function")
def eurosat_data_root(tmp_path):
    data_root = tmp_path / "m_eurosat"
    data_directory = data_root / MEuroSATNonGeo.data_dir
    data_directory.mkdir(parents=True, exist_ok=True)

    partition = {
        "train": ["image_0", "image_1"],
        "val": [],
        "test": []
    }
    partition_file = data_directory / MEuroSATNonGeo.partition_file_template.format(partition="default")
    with open(partition_file, "w") as f:
        json.dump(partition, f)

    label_map = {
        "class_0": ["image_0"],
        "class_1": ["image_1"]
    }
    label_map_file = data_directory / MEuroSATNonGeo.label_map_file
    with open(label_map_file, "w") as f:
        json.dump(label_map, f)

    bands = [
        "COASTAL_AEROSOL", "BLUE", "GREEN", "RED", "RED_EDGE_1", "RED_EDGE_2", "RED_EDGE_3",
        "NIR_BROAD", "NIR_NARROW", "WATER_VAPOR", "CIRRUS", "SWIR_1", "SWIR_2"
    ]

    for img in ["image_0", "image_1"]:
        file_path = data_directory / f"{img}.hdf5"
        with h5py.File(file_path, "w") as h5file:
            for band in bands:
                h5file.create_dataset(band, data=np.random.rand(100, 100).astype(np.float32))

            label_tensor = torch.tensor(0 if img == "image_0" else 1, dtype=torch.long)
            h5file.create_dataset("label", data=label_tensor.numpy())

    return str(data_root)

@pytest.fixture(scope="function")
def pv4gerseg_data_root(tmp_path):
    data_root = tmp_path / "m_pv4ger_seg"
    data_directory = data_root / MPv4gerSegNonGeo.data_dir
    data_directory.mkdir(parents=True, exist_ok=True)

    partition = {
        "train": ["52.5167,13.3833", "48.8566,2.3522"],
        "val": [],
        "test": []
    }
    partition_file = data_directory / MPv4gerSegNonGeo.partition_file_template.format(partition="default")
    with open(partition_file, "w") as f:
        json.dump(partition, f)

    bands = ["BLUE", "GREEN", "RED"]

    for img in ["52.5167,13.3833", "48.8566,2.3522"]:
        file_path = data_directory / f"{img}.hdf5"
        with h5py.File(file_path, "w") as h5file:
            for band in bands:
                h5file.create_dataset(band, data=np.random.rand(100, 100).astype(np.float32))

            mask = np.random.randint(0, 3, size=(100, 100), dtype=np.uint8)
            h5file.create_dataset("label", data=mask)

    return str(data_root)

@pytest.fixture(scope="function")
def pv4ger_data_root(tmp_path):
    data_root = tmp_path / "m_pv4ger"
    data_directory = data_root / MPv4gerNonGeo.data_dir
    data_directory.mkdir(parents=True, exist_ok=True)

    partition = {
        "train": ["52.5167,13.3833", "48.8566,2.3522"],
        "val": [],
        "test": []
    }
    partition_file = data_directory / MPv4gerNonGeo.partition_file_template.format(partition="default")
    with open(partition_file, "w") as f:
        json.dump(partition, f)

    bands = ["BLUE", "GREEN", "RED"]

    for img in ["52.5167,13.3833", "48.8566,2.3522"]:
        file_path = data_directory / f"{img}.hdf5"
        with h5py.File(file_path, "w") as h5file:
            for band in bands:
                h5file.create_dataset(band, data=np.random.rand(100, 100).astype(np.float32))

            attr_dict = {"label": 1 if img == "52.5167,13.3833" else 0}
            h5file.attrs["pickle"] = str(pickle.dumps(attr_dict))

    return str(data_root)

@pytest.fixture(scope="function")
def so2sat_data_root(tmp_path):
    data_root = tmp_path / "m_so2sat"
    data_directory = data_root / MSo2SatNonGeo.data_dir
    data_directory.mkdir(parents=True, exist_ok=True)

    partition = {
        "train": ["image_0", "image_1"],
        "val": [],
        "test": []
    }
    partition_file = data_directory / MSo2SatNonGeo.partition_file_template.format(partition="default")
    with open(partition_file, "w") as f:
        json.dump(partition, f)

    bands = [
        "VH_REAL", "BLUE", "VH_IMAGINARY", "GREEN", "VV_REAL", "RED",
        "VV_IMAGINARY", "VH_LEE_FILTERED", "RED_EDGE_1", "VV_LEE_FILTERED",
        "RED_EDGE_2", "VH_LEE_FILTERED_REAL", "RED_EDGE_3", "NIR_BROAD",
        "VV_LEE_FILTERED_IMAGINARY", "NIR_NARROW", "SWIR_1", "SWIR_2"
    ]

    for img in ["image_0", "image_1"]:
        file_path = data_directory / f"{img}.hdf5"
        with h5py.File(file_path, "w") as h5file:
            for band in bands:
                h5file.create_dataset(band, data=np.random.rand(100, 100).astype(np.float32))

            attr_dict = {"label": 1 if img == "image_0" else 0}
            h5file.attrs["pickle"] = str(pickle.dumps(attr_dict))

    return str(data_root)

@pytest.fixture(scope="function")
def cashews_data_root(tmp_path):
    data_root = tmp_path / "m_cashew_plant"
    data_directory = data_root / MBeninSmallHolderCashewsNonGeo.data_dir
    data_directory.mkdir(parents=True, exist_ok=True)

    partition = {
        "train": ["image_2023-01-15", "image_2023-02-20"],
        "val": [],
        "test": []
    }
    partition_file = data_directory / MBeninSmallHolderCashewsNonGeo.partition_file_template.format(partition="default")
    with open(partition_file, "w") as f:
        json.dump(partition, f)

    bands = [
        "COASTAL_AEROSOL", "BLUE", "GREEN", "RED", "RED_EDGE_1", "RED_EDGE_2",
        "RED_EDGE_3", "NIR_BROAD", "NIR_NARROW", "WATER_VAPOR", "SWIR_1", "SWIR_2", "CLOUD_PROBABILITY"
    ]

    for img in ["image_2023-01-15", "image_2023-02-20"]:
        file_path = data_directory / f"{img}.hdf5"
        with h5py.File(file_path, "w") as h5file:
            for band in bands:
                date_band_name = f"{band}_2023-01-15" if img == "image_2023-01-15" else f"{band}_2023-02-20"
                h5file.create_dataset(date_band_name, data=np.random.rand(100, 100).astype(np.float32))

            mask = np.random.randint(0, 3, size=(100, 100), dtype=np.uint8)
            h5file.create_dataset("label", data=mask)

    return str(data_root)

@pytest.fixture(scope="function")
def sen1floods_data_root(tmp_path):
    data_root = tmp_path / "sen1floods11"
    (data_root / Sen1Floods11NonGeo.data_dir).mkdir(parents=True, exist_ok=True)
    (data_root / Sen1Floods11NonGeo.label_dir).mkdir(parents=True, exist_ok=True)
    (data_root / Sen1Floods11NonGeo.split_dir).mkdir(parents=True, exist_ok=True)

    for i in range(5):
        filename = f"tile_{i}_S2Hand.tif"
        label_filename = f"tile_{i}_LabelHand.tif"

        img_data = DataArray(
            np.random.rand(13, 64, 64).astype(np.float32),
            dims=["band", "y", "x"]
        )
        mask_data = DataArray(
            np.random.randint(0, 2, size=(1, 64, 64), dtype=np.uint8),
            dims=["band", "y", "x"]
        )

        img_data.rio.set_spatial_dims(x_dim="x", y_dim="y", inplace=True)
        mask_data.rio.set_spatial_dims(x_dim="x", y_dim="y", inplace=True)

        image_path = data_root / Sen1Floods11NonGeo.data_dir / filename
        mask_path = data_root / Sen1Floods11NonGeo.label_dir / label_filename

        img_data.rio.to_raster(str(image_path))
        mask_data.rio.to_raster(str(mask_path))

    split_file = data_root / Sen1Floods11NonGeo.split_dir / "flood_train_data.txt"
    with open(split_file, "w") as f:
        f.write("\n".join([f"tile_{i}" for i in range(5)]))

    metadata = {
        "type": "FeatureCollection",
        "features": [
            {
                "type": "Feature",
                "properties": {"location": f"tile_{i}", "s2_date": "2021-01-01"},
                "geometry": {"type": "Point", "coordinates": [0, 0]},
            }
            for i in range(5)
        ],
    }
    metadata_path = data_root / Sen1Floods11NonGeo.metadata_file
    with open(metadata_path, "w") as f:
        json.dump(metadata, f)

    return str(data_root)

@pytest.fixture(scope="function")
def chesapeake_data_root(tmp_path):
    data_root = tmp_path / "m_chesapeake"
    data_directory = data_root / MChesapeakeLandcoverNonGeo.data_dir
    data_directory.mkdir(parents=True, exist_ok=True)

    partition = {
        "train": ["image_0", "image_1"],
        "val": [],
        "test": []
    }
    partition_file = data_directory / MChesapeakeLandcoverNonGeo.partition_file_template.format(partition="default")
    with open(partition_file, "w") as f:
        json.dump(partition, f)

    bands = ["BLUE", "GREEN", "NIR", "RED"]

    for img in ["image_0", "image_1"]:
        file_path = data_directory / f"{img}.hdf5"
        with h5py.File(file_path, "w") as h5file:
            for band in bands:
                h5file.create_dataset(band, data=np.random.rand(100, 100).astype(np.float32))

            mask = np.random.randint(0, 3, size=(100, 100), dtype=np.uint8)
            h5file.create_dataset("label", data=mask)

    return str(data_root)

@pytest.fixture(scope="function")
def crop_classification_data_root(tmp_path):
    data_root = tmp_path / "crop_classification"
    training_dir = data_root / "training_chips"
    validation_dir = data_root / "validation_chips"

    training_dir.mkdir(parents=True, exist_ok=True)
    validation_dir.mkdir(parents=True, exist_ok=True)

    for directory in [training_dir, validation_dir]:
        for i in range(2):
            filename = f"chip_{i}_merged.tif"
            label_filename = f"chip_{i}.mask.tif"
            img_data = DataArray(np.random.rand(18, 64, 64).astype(np.float32), dims=["band", "y", "x"])
            mask_data = DataArray(np.random.randint(0, 13, size=(1, 64, 64), dtype=np.uint8), dims=["band", "y", "x"])

            img_data.rio.set_spatial_dims(x_dim="x", y_dim="y", inplace=True)
            mask_data.rio.set_spatial_dims(x_dim="x", y_dim="y", inplace=True)
            img_data.rio.write_crs("EPSG:4326", inplace=True)
            mask_data.rio.write_crs("EPSG:4326", inplace=True)
            image_path = directory / filename
            mask_path = directory / label_filename
            img_data.rio.to_raster(str(image_path))
            mask_data.rio.to_raster(str(mask_path))

    with open(data_root / "training_data.txt", "w") as f:
        f.write("\n".join([f"chip_{i}" for i in range(2)]))

    with open(data_root / "validation_data.txt", "w") as f:
        f.write("\n".join([f"chip_{i}" for i in range(2)]))

    metadata = pd.DataFrame({
        "chip_id": [f"chip_{i}" for i in range(2)],
        "first_img_date": ["2021-01-01", "2021-01-02"],
        "middle_img_date": ["2021-01-15", "2021-01-16"],
        "last_img_date": ["2021-02-01", "2021-02-02"],
    })
    metadata.to_csv(data_root / "chips_df.csv", index=False)

    return str(data_root)

class TestMNeonTreeNonGeo:
    def test_dataset_sample(self, neontree_data_root):
        transform = A.Compose([
            A.Resize(64, 64),
            ToTensorV2(),
        ])

        dataset = MNeonTreeNonGeo(data_root=neontree_data_root, split="train", transform=transform)
        sample = dataset[0]

        assert "image" in sample, "Sample does not contain 'image'"
        assert "mask" in sample, "Sample does not contain 'mask'"
        assert isinstance(sample["image"], torch.Tensor), "'image' is not a torch.Tensor"
        assert isinstance(sample["mask"], torch.Tensor), "'mask' is not a torch.Tensor"
        assert sample["image"].dtype == torch.float32, "'image' does not have dtype torch.float32"
        assert sample["mask"].dtype == torch.long, "'mask' does not have dtype torch.long"
        assert sample["image"].shape == (len(dataset.bands), 64, 64), f"'image' has incorrect shape: {sample['image'].shape}"
        assert sample["mask"].shape == (64, 64), f"'mask' has incorrect shape: {sample['mask'].shape}"

    def test_plot(self, neontree_data_root):
        transform = A.Compose([ToTensorV2()])

        dataset = MNeonTreeNonGeo(data_root=neontree_data_root, split="train", transform=transform)
        sample = dataset[0]

        fig = dataset.plot(sample, suptitle="Sample Plot")
        assert isinstance(fig, plt.Figure), "Plot method did not return a plt.Figure"
        plt.close(fig)


class TestMBrickKilnNonGeo:
    def test_dataset_sample(self, brickkiln_data_root):
        transform = A.Compose([
            A.Resize(64, 64),
            ToTensorV2(),
        ])

        dataset = MBrickKilnNonGeo(data_root=brickkiln_data_root, split="train", transform=transform)
        sample = dataset[0]

        assert "image" in sample, "Sample does not contain 'image'"
        assert "label" in sample, "Sample does not contain 'label'"
        assert isinstance(sample["image"], torch.Tensor), "'image' is not a torch.Tensor"
        assert isinstance(sample["label"], int), "'label' is not an int"
        assert sample["image"].dtype == torch.float32, "'image' does not have dtype torch.float32"
        assert sample["image"].shape == (len(dataset.bands), 64, 64), f"'image' has incorrect shape: {sample['image'].shape}"

    def test_plot(self, brickkiln_data_root):
        transform = A.Compose([ToTensorV2()])

        dataset = MBrickKilnNonGeo(data_root=brickkiln_data_root, split="train", transform=transform)
        sample = dataset[0]

        fig = dataset.plot(sample, suptitle="Sample Plot")
        assert isinstance(fig, plt.Figure), "Plot method did not return a plt.Figure"
        plt.close(fig)

class TestMEuroSATNonGeo:
    def test_dataset_sample(self, eurosat_data_root):
        transform = A.Compose([
            A.Resize(64, 64),
            ToTensorV2(),
        ])

        dataset = MEuroSATNonGeo(data_root=eurosat_data_root, split="train", transform=transform)
        sample = dataset[0]

        assert "image" in sample, "Sample does not contain 'image'"
        assert "label" in sample, "Sample does not contain 'label'"
        assert isinstance(sample["image"], torch.Tensor), "'image' is not a torch.Tensor"
        assert isinstance(sample["label"], int), "'label' is not an int"
        assert sample["image"].dtype == torch.float32, "'image' does not have dtype torch.float32"
        assert sample["image"].shape == (len(dataset.bands), 64, 64), f"'image' has incorrect shape: {sample['image'].shape}"

    def test_plot(self, eurosat_data_root):
        transform = A.Compose([ToTensorV2()])

        dataset = MEuroSATNonGeo(data_root=eurosat_data_root, split="train", transform=transform)
        sample = dataset[0]

        fig = dataset.plot(sample, suptitle="Sample Plot")
        assert isinstance(fig, plt.Figure), "Plot method did not return a plt.Figure"
        plt.close(fig)

class TestFireScarsNonGeo:
    def test_dataset_length(self, fire_scars_data_root):
        dataset = FireScarsNonGeo(data_root=fire_scars_data_root, split="train")
        expected_length = 5
        actual_length = len(dataset)
        assert actual_length == expected_length, f"Expected {expected_length}, but got {actual_length}"

    def test_dataset_sample(self, fire_scars_data_root):
        dataset = FireScarsNonGeo(data_root=fire_scars_data_root, split="train")
        sample = dataset[0]

        assert "image" in sample, "Sample does not contain 'image'"
        assert "mask" in sample, "Sample does not contain 'mask'"
        assert isinstance(sample["image"], torch.Tensor), "'image' is not a torch.Tensor"
        assert isinstance(sample["mask"], torch.Tensor), "'mask' is not a torch.Tensor"
        assert sample["image"].dtype == torch.float32, "'image' does not have dtype torch.float32"
        assert sample["mask"].dtype in [torch.float32, torch.long], "'mask' does not have expected dtype (torch.float32 or torch.long)"
        assert sample["image"].ndim == 3, "'image' does not have 3 dimensions (C, H, W)"
        assert sample["mask"].ndim == 2, "'mask' does not have 2 dimensions (H, W)"

    def test_plot(self, fire_scars_data_root):
        transform = A.Compose([ToTensorV2()])

        dataset = FireScarsNonGeo(
            data_root=fire_scars_data_root,
            split="train",
            transform=transform,
        )
        sample = dataset[0]

        fig = dataset.plot(sample, suptitle="Sample Plot")
        assert isinstance(fig, plt.Figure), "The plot method did not return a plt.Figure"
        plt.close(fig)

class TestMBigEarthNonGeo:
    def test_dataset_sample(self, m_bigearth_data_root):
        transform = A.Compose([
            ToTensorV2(),
        ])

        dataset = MBigEarthNonGeo(data_root=m_bigearth_data_root, split="train", transform=transform)
        sample = dataset[0]

        assert "image" in sample, "Sample does not contain 'image'"
        assert "label" in sample, "Sample does not contain 'label'"
        assert isinstance(sample["image"], torch.Tensor), "'image' is not a torch.Tensor"
        assert isinstance(sample["label"], torch.Tensor), "'label' is not a torch.Tensor"
        assert sample["image"].dtype == torch.float32, "'image' does not have dtype torch.float32"
        assert sample["label"].dtype == torch.float32, "'label' does not have dtype torch.float32"

    def test_plot(self, m_bigearth_data_root):
        transform = A.Compose([ToTensorV2()])

        dataset = MBigEarthNonGeo(
            data_root=m_bigearth_data_root,
            split="train",
            transform=transform,
        )
        sample = dataset[0]

        fig = dataset.plot(sample, suptitle="Sample Plot")
        assert isinstance(fig, plt.Figure), "The plot method did not return a plt.Figure"
        plt.close(fig)

class TestMForestNetNonGeo:
    def test_dataset_sample(self, m_forestnet_data_root):
        transform = A.Compose([
            ToTensorV2(),
        ])

        dataset = MForestNetNonGeo(data_root=m_forestnet_data_root, split="train", transform=transform)
        sample = dataset[0]

        assert "image" in sample, "Sample does not contain 'image'"
        assert "label" in sample, "Sample does not contain 'label'"
        assert isinstance(sample["image"], torch.Tensor), "'image' is not a torch.Tensor"
        assert isinstance(sample["label"], int), "'label' is not an int"
        assert sample["image"].dtype == torch.float32, "'image' does not have dtype torch.float32"
        assert isinstance(sample["label"], int), "'label' is not an int"

    def test_plot(self, m_forestnet_data_root):
        transform = A.Compose([ToTensorV2()])

        dataset = MForestNetNonGeo(
            data_root=m_forestnet_data_root,
            split="train",
            transform=transform,
        )
        sample = dataset[0]

        fig = dataset.plot(sample, suptitle="Sample Plot")
        assert isinstance(fig, plt.Figure), "The plot method did not return a plt.Figure"
        plt.close(fig)

class TestMNzCattleNonGeo:
    def test_dataset_sample(self, mnz_cattle_data_root):
        transform = A.Compose([
            A.Resize(64, 64),
            ToTensorV2(),
        ])

        dataset = MNzCattleNonGeo(data_root=mnz_cattle_data_root, split="train", transform=transform)
        sample = dataset[0]

        assert "image" in sample, "Sample does not contain 'image'"
        assert "mask" in sample, "Sample does not contain 'mask'"
        assert isinstance(sample["image"], torch.Tensor), "'image' is not a torch.Tensor"
        assert isinstance(sample["mask"], torch.Tensor), "'mask' is not a torch.Tensor"
        assert sample["image"].dtype == torch.float32, "'image' does not have dtype torch.float32"
        assert sample["mask"].dtype == torch.long, "'mask' does not have dtype torch.long"
        assert sample["image"].shape == (3, 64, 64), f"'image' has incorrect shape: {sample['image'].shape}"
        assert sample["mask"].shape == (64, 64), f"'mask' has incorrect shape: {sample['mask'].shape}"

    def test_plot(self, mnz_cattle_data_root):
        transform = A.Compose([ToTensorV2()])

        dataset = MNzCattleNonGeo(
            data_root=mnz_cattle_data_root,
            split="train",
            transform=transform,
        )
        sample = dataset[0]

        fig = dataset.plot(sample, suptitle="Sample Plot")
        assert isinstance(fig, plt.Figure), "The plot method did not return a plt.Figure"
        plt.close(fig)

class TestMPv4gerNonGeo:
    def test_dataset_sample(self, pv4ger_data_root):
        transform = A.Compose([
            A.Resize(64, 64),
            ToTensorV2(),
        ])

        dataset = MPv4gerNonGeo(data_root=pv4ger_data_root, split="train", transform=transform, use_metadata=True)
        sample = dataset[0]
        assert "image" in sample, "Sample does not contain 'image'"
        assert "label" in sample, "Sample does not contain 'label'"
        assert "location_coords" in sample, "Sample does not contain 'location_coords'"
        assert isinstance(sample["image"], torch.Tensor), "'image' is not a torch.Tensor"
        assert isinstance(sample["label"], int), "'label' is not an int"
        assert isinstance(sample["location_coords"], torch.Tensor), "'location_coords' is not a torch.Tensor"
        assert sample["image"].dtype == torch.float32, "'image' does not have dtype torch.float32"
        assert sample["location_coords"].dtype == torch.float32, "'location_coords' does not have dtype torch.float32"
        assert sample["image"].shape == (len(dataset.bands), 64, 64), f"'image' has incorrect shape: {sample['image'].shape}"
        assert sample["location_coords"].shape == (2,), f"'location_coords' has incorrect shape: {sample['location_coords'].shape}"

    def test_plot(self, pv4ger_data_root):

        transform = A.Compose([ToTensorV2()])


        dataset = MPv4gerNonGeo(data_root=pv4ger_data_root, split="train", transform=transform)
        sample = dataset[0]


        fig = dataset.plot(sample, suptitle="Sample Plot")
        assert isinstance(fig, plt.Figure), "Plot method did not return a plt.Figure"
        plt.close(fig)


class TestMPv4gerSegNonGeo:
    def test_dataset_sample(self, pv4gerseg_data_root):

        transform = A.Compose([
            A.Resize(64, 64),
            ToTensorV2(),
        ])

        dataset = MPv4gerSegNonGeo(data_root=pv4gerseg_data_root, split="train", transform=transform, use_metadata=True)
        sample = dataset[0]

        assert "image" in sample, "Sample does not contain 'image'"
        assert "mask" in sample, "Sample does not contain 'mask'"
        assert "location_coords" in sample, "Sample does not contain 'location_coords'"
        assert isinstance(sample["image"], torch.Tensor), "'image' is not a torch.Tensor"
        assert isinstance(sample["mask"], torch.Tensor), "'mask' is not a torch.Tensor"
        assert isinstance(sample["location_coords"], torch.Tensor), "'location_coords' is not a torch.Tensor"
        assert sample["image"].dtype == torch.float32, "'image' does not have dtype torch.float32"
        assert sample["mask"].dtype == torch.long, "'mask' does not have dtype torch.long"
        assert sample["location_coords"].dtype == torch.float32, "'location_coords' does not have dtype torch.float32"
        assert sample["image"].shape == (len(dataset.bands), 64, 64), f"'image' has incorrect shape: {sample['image'].shape}"
        assert sample["mask"].shape == (64, 64), f"'mask' has incorrect shape: {sample['mask'].shape}"
        assert sample["location_coords"].shape == (2,), f"'location_coords' has incorrect shape: {sample['location_coords'].shape}"

    def test_plot(self, pv4gerseg_data_root):
        transform = A.Compose([ToTensorV2()])

        dataset = MPv4gerSegNonGeo(data_root=pv4gerseg_data_root, split="train", transform=transform)
        sample = dataset[0]

        fig = dataset.plot(sample, suptitle="Sample Plot")
        assert isinstance(fig, plt.Figure), "Plot method did not return a plt.Figure"
        plt.close(fig)

@pytest.fixture(scope="function")
def sacroptype_data_root(tmp_path):

    data_root = tmp_path / "m_sa_crop_type"
    data_directory = data_root / MSACropTypeNonGeo.data_dir
    data_directory.mkdir(parents=True, exist_ok=True)


    partition = {
        "train": ["image_0", "image_1"],
        "val": [],
        "test": []
    }
    partition_file = data_directory / MSACropTypeNonGeo.partition_file_template.format(partition="default")
    with open(partition_file, "w") as f:
        json.dump(partition, f)


    bands = [
        "COASTAL_AEROSOL", "BLUE", "GREEN", "RED", "RED_EDGE_1", "RED_EDGE_2",
        "RED_EDGE_3", "NIR_BROAD", "NIR_NARROW", "WATER_VAPOR", "SWIR_1", "SWIR_2", "CLOUD_PROBABILITY"
    ]


    for img in ["image_0", "image_1"]:
        file_path = data_directory / f"{img}.hdf5"
        with h5py.File(file_path, "w") as h5file:
            for band in bands:
                h5file.create_dataset(band, data=np.random.rand(100, 100).astype(np.float32))

            mask = np.random.randint(0, 3, size=(100, 100), dtype=np.uint8)
            h5file.create_dataset("label", data=mask)

    return str(data_root)


class TestMSACropTypeNonGeo:
    def test_dataset_sample(self, sacroptype_data_root):

        transform = A.Compose([
            A.Resize(64, 64),
            ToTensorV2(),
        ])

        dataset = MSACropTypeNonGeo(data_root=sacroptype_data_root, split="train", transform=transform)
        sample = dataset[0]

        assert "image" in sample, "Sample does not contain 'image'"
        assert "mask" in sample, "Sample does not contain 'mask'"
        assert isinstance(sample["image"], torch.Tensor), "'image' is not a torch.Tensor"
        assert isinstance(sample["mask"], torch.Tensor), "'mask' is not a torch.Tensor"
        assert sample["image"].dtype == torch.float32, "'image' does not have dtype torch.float32"
        assert sample["mask"].dtype == torch.long, "'mask' does not have dtype torch.long"
        assert sample["image"].shape == (len(dataset.bands), 64, 64), f"'image' has incorrect shape: {sample['image'].shape}"
        assert sample["mask"].shape == (64, 64), f"'mask' has incorrect shape: {sample['mask'].shape}"

    def test_plot(self, sacroptype_data_root):
        transform = A.Compose([ToTensorV2()])

        dataset = MSACropTypeNonGeo(data_root=sacroptype_data_root, split="train", transform=transform)
        sample = dataset[0]

        fig = dataset.plot(sample, suptitle="Sample Plot")
        assert isinstance(fig, plt.Figure), "Plot method did not return a plt.Figure"
        plt.close(fig)


class TestMSo2SatNonGeo:
    def test_dataset_sample(self, so2sat_data_root):
        transform = A.Compose([
            A.Resize(64, 64),
            ToTensorV2(),
        ])

        dataset = MSo2SatNonGeo(data_root=so2sat_data_root, split="train", transform=transform)
        sample = dataset[0]

        assert "image" in sample, "Sample does not contain 'image'"
        assert "label" in sample, "Sample does not contain 'label'"
        assert isinstance(sample["image"], torch.Tensor), "'image' is not a torch.Tensor"
        assert isinstance(sample["label"], int), "'label' is not an int"
        assert sample["image"].dtype == torch.float32, "'image' does not have dtype torch.float32"
        assert sample["image"].shape == (len(dataset.bands), 64, 64), f"'image' has incorrect shape: {sample['image'].shape}"

    def test_plot(self, so2sat_data_root):
        transform = A.Compose([ToTensorV2()])

        dataset = MSo2SatNonGeo(data_root=so2sat_data_root, split="train", transform=transform)
        sample = dataset[0]

        fig = dataset.plot(sample, suptitle="Sample Plot")
        assert isinstance(fig, plt.Figure), "Plot method did not return a plt.Figure"
        plt.close(fig)

class TestMBeninSmallHolderCashewsNonGeo:
    def test_dataset_sample(self, cashews_data_root):
        transform = A.Compose([
            A.Resize(64, 64),
            ToTensorV2(),
        ])

        dataset = MBeninSmallHolderCashewsNonGeo(data_root=cashews_data_root, split="train", transform=transform, use_metadata=True)
        sample = dataset[0]

        assert "image" in sample, "Sample does not contain 'image'"
        assert "mask" in sample, "Sample does not contain 'mask'"
        assert "temporal_coords" in sample, "Sample does not contain 'temporal_coords'"
        assert isinstance(sample["image"], torch.Tensor), "'image' is not a torch.Tensor"
        assert isinstance(sample["mask"], torch.Tensor), "'mask' is not a torch.Tensor"
        assert isinstance(sample["temporal_coords"], torch.Tensor), "'temporal_coords' is not a torch.Tensor"
        assert sample["image"].dtype == torch.float32, "'image' does not have dtype torch.float32"
        assert sample["mask"].dtype == torch.long, "'mask' does not have dtype torch.long"
        assert sample["temporal_coords"].dtype == torch.float32, "'temporal_coords' does not have dtype torch.float32"
        assert sample["image"].shape == (len(dataset.bands), 64, 64), f"'image' has incorrect shape: {sample['image'].shape}"
        assert sample["mask"].shape == (64, 64), f"'mask' has incorrect shape: {sample['mask'].shape}"
        assert sample["temporal_coords"].shape == (1, 2), f"'temporal_coords' has incorrect shape: {sample['temporal_coords'].shape}"

    def test_plot(self, cashews_data_root):
        transform = A.Compose([ToTensorV2()])

        dataset = MBeninSmallHolderCashewsNonGeo(data_root=cashews_data_root, split="train", transform=transform)
        sample = dataset[0]

        fig = dataset.plot(sample, suptitle="Sample Plot")
        assert isinstance(fig, plt.Figure), "Plot method did not return a plt.Figure"
        plt.close(fig)


class TestMChesapeakeLandcoverNonGeo:
    def test_dataset_sample(self, chesapeake_data_root):

        transform = A.Compose([
            A.Resize(64, 64),
            ToTensorV2(),
        ])

        dataset = MChesapeakeLandcoverNonGeo(data_root=chesapeake_data_root, split="train", transform=transform)
        sample = dataset[0]

        assert "image" in sample, "Sample does not contain 'image'"
        assert "mask" in sample, "Sample does not contain 'mask'"
        assert isinstance(sample["image"], torch.Tensor), "'image' is not a torch.Tensor"
        assert isinstance(sample["mask"], torch.Tensor), "'mask' is not a torch.Tensor"
        assert sample["image"].dtype == torch.float32, "'image' does not have dtype torch.float32"
        assert sample["mask"].dtype == torch.long, "'mask' does not have dtype torch.long"
        assert sample["image"].shape == (len(dataset.bands), 64, 64), f"'image' has incorrect shape: {sample['image'].shape}"
        assert sample["mask"].shape == (64, 64), f"'mask' has incorrect shape: {sample['mask'].shape}"

    def test_plot(self, chesapeake_data_root):
        transform = A.Compose([ToTensorV2()])

        dataset = MChesapeakeLandcoverNonGeo(data_root=chesapeake_data_root, split="train", transform=transform)
        sample = dataset[0]

        fig = dataset.plot(sample, suptitle="Sample Plot")
        assert isinstance(fig, plt.Figure), "Plot method did not return a plt.Figure"
        plt.close(fig)

class TestSen1Floods11NonGeo:
    def test_dataset_sample(self, sen1floods_data_root):
        transform = A.Compose([
            A.Resize(32, 32),
            ToTensorV2(),
        ])

        dataset = Sen1Floods11NonGeo(data_root=sen1floods_data_root, split="train", transform=transform, use_metadata=True)
        sample = dataset[0]

        assert "image" in sample, "Sample does not contain 'image'"
        assert "mask" in sample, "Sample does not contain 'mask'"
        assert "location_coords" in sample, "Sample does not contain 'location_coords'"
        assert "temporal_coords" in sample, "Sample does not contain 'temporal_coords'"
        assert isinstance(sample["image"], torch.Tensor), "'image' is not a torch.Tensor"
        assert isinstance(sample["mask"], torch.Tensor), "'mask' is not a torch.Tensor"
        assert isinstance(sample["location_coords"], torch.Tensor), "'location_coords' is not a torch.Tensor"
        assert isinstance(sample["temporal_coords"], torch.Tensor), "'temporal_coords' is not a torch.Tensor"
        assert sample["image"].dtype == torch.float32, "'image' does not have dtype torch.float32"
        assert sample["mask"].dtype == torch.long, "'mask' does not have dtype torch.long"
        assert sample["location_coords"].dtype == torch.float32, "'location_coords' does not have dtype torch.float32"
        assert sample["temporal_coords"].dtype == torch.float32, "'temporal_coords' does not have dtype torch.float32"
        num_bands = len(dataset.bands)
        assert sample["image"].shape == (num_bands, 32, 32), f"'image' has incorrect shape: {sample['image'].shape}"
        assert sample["mask"].shape == (32, 32), f"'mask' has incorrect shape: {sample['mask'].shape}"

    def test_plot(self, sen1floods_data_root):
        transform = A.Compose([ToTensorV2()])

        dataset = Sen1Floods11NonGeo(data_root=sen1floods_data_root, split="train", transform=transform)
        sample = dataset[0]

        fig = dataset.plot(sample, suptitle="Sample Plot")
        assert isinstance(fig, plt.Figure), "The plot method did not return a plt.Figure"
        plt.close(fig)

class TestMultiTemporalCropClassification:
    def test_dataset_sample(self, crop_classification_data_root):
        transform = A.Compose(
            [
                FlattenTemporalIntoChannels(),
                ToTensorV2(),
                UnflattenTemporalFromChannels(3),
            ],
            is_check_shapes=False
        )

        dataset = MultiTemporalCropClassification(
            data_root=crop_classification_data_root,
            split="train",
            transform=transform,
            use_metadata=True,
        )
        sample = dataset[0]

        assert "image" in sample, "Sample does not contain 'image'"
        assert "mask" in sample, "Sample does not contain 'mask'"
        assert isinstance(sample["image"], torch.Tensor), "'image' is not a torch.Tensor"
        assert isinstance(sample["mask"], torch.Tensor), "'mask' is not a torch.Tensor"
        assert sample["image"].dtype == torch.float32, "'image' does not have dtype torch.float32"
        assert sample["mask"].dtype == torch.long, "'mask' does not have dtype torch.long"
        assert sample["image"].shape == (6, 3, 64, 64), f"'image' has incorrect shape: {sample['image'].shape}"
        assert sample["mask"].shape == (64, 64), f"'mask' has incorrect shape: {sample['mask'].shape}"

    def test_plot(self, crop_classification_data_root):
        transform = A.Compose(
            [
                FlattenTemporalIntoChannels(),
                ToTensorV2(),
                UnflattenTemporalFromChannels(3),
            ],
            is_check_shapes=False
        )

        dataset = MultiTemporalCropClassification(
            data_root=crop_classification_data_root,
            split="train",
            transform=transform,
        )
        sample = dataset[0]

        fig = dataset.plot(sample, suptitle="Sample Plot")
        assert isinstance(fig, plt.Figure), "The plot method did not return a plt.Figure"
        plt.close(fig)
